# Standard library imports
import warnings

# Third party imports
import numpy as np
from pyomo.environ import Constraint, Objective, Param, Set, SolverFactory, Suffix, value, Var

# KIPET library imports
from kipet.estimator_tools.parameter_estimator import ParameterEstimator


class EstimabilityAnalyzer(ParameterEstimator):
    """This class is for estimability analysis. For now it will be used to select the parameter set that
    is suitable for estimation based on a mean squared error (MSE) approach first described by Wu, McLean,
    Harris, and McAuley (2011). This, in time, will be expanded to be able to do estimability analysis 
    for spectral data problems as well. The class will contain a number of functions that will perform the 
    estimability analysis. 

    Parameters
    ----------
    :param ConcreteModel model: The full model TemplateBuilder problem needs to be fed into the Estimability Analyzer
        as this is needed in order to build the sensitivities for ranking parameters as well as for constructing the
        simplified models

    """
    def __init__(self, model):
        """Initialization of the EstimabilityAnalyzer"""
        super(EstimabilityAnalyzer, self).__init__(model)
        self.param_ranks = dict()
        
    def get_sensitivities_for_params(self, **kwds):
        """ Obtains the sensitivities (dsdp) using k_aug. This function only works for
        concentration-only problems and obtains the sensitivities based on the initial parameter
        values and how they affect the output (Z).

        :Keyword Args:

            - sigmasq (dict): map of component name to noise variance. The map also contains the device noise variance
            - tee (bool,optional): flag to tell the optimizer whether to stream output to the terminal or not
            
        :return numpy.ndarray dsdp: Sensitivity matrix with columns being parameters and rows the Z vars
        :return dict idx_to_params: Dictionary that maps the columns to the parameters

        """               
        if not self.model.alltime.get_discretization_info():
            raise RuntimeError('apply discretization first before running the estimability')
            
        sigma_sq = kwds.pop('sigmasq', dict())
        tee = kwds.pop('tee', False)
        species_list = kwds.pop('subset_components', None)

        list_components = []
        if species_list is None:
            list_components = [k for k in self._mixture_components]
        else:
            for k in species_list:
                if k in self._mixture_components:
                    list_components.append(k)
                else:
                    warnings.warn("Ignored {} since is not a mixture component of the model".format(k))

        if not self._concentration_given:
            raise NotImplementedError("In order to use the estimability analysis from concentration data requires concentration data model.C[ti,cj]")

        if self._huplc_given:
            raise NotImplementedError("Estimability analysis for additional huplc data is not implemented yet.")

        all_sigma_specified = True

        keys = sigma_sq.keys()
        for k in list_components:
            if k not in keys:
                all_sigma_specified = False
                sigma_sq[k] = max(sigma_sq.values())
                
        if not all_sigma_specified:
            raise RuntimeError(
                'All variances must be specified to determine sensitivities.\n Please pass variance dictionary to run_opt')
        
        m = self.model

        # estimation
        def rule_objective(m):
            obj = 0
            for t in m.allmeas_times:
                obj += sum((m.Cm[t, k] - m.Z[t, k]) ** 2 / sigma_sq[k] for k in list_components)
            return obj
            
        m.objective = Objective(rule=rule_objective)

        #set dummy variables for k_aug to do sensitivities
        paramcount = 0
        paramlist=list()
        varcount = 0
        varlist = list()
        for k,v in m.P.items():
            if v.is_fixed():
                paramcount +=1
                paramlist.append(k)
            else:
                varcount += 1
                varlist.append(k)
                
        if paramcount >= 1:
            m.dpset = Set(initialize = paramlist)
            m.dummyP= Var(m.dpset)
            for i in paramlist:
                #print("dummyP", i)
                m.dummyP[i] = m.P[i].value
                
        if varcount >= 1:
            m.dvset = Set(initialize = varlist)
            m.dummyV= Param(m.dvset, mutable =True)
            for i in varlist:
                #print("dummyV", i)
                #print(m.P[i].value)
                m.dummyV[i] = m.P[i].value
        
        #set dummy constraints   
        def dummy_constraints(m,p):
            if p in varlist:
                return 0 == m.dummyV[p] - m.P[p] 
            if p in paramlist:
                return 0 == m.dummyP[p] - m.P[p] 
           
        m.dummyC = Constraint(m.parameter_names, rule=dummy_constraints)     
        
        
        if hasattr(m, 'dual'):
            print('The model has duals already...')
        #set up suffixes for Ipopt that are required for k_aug
        m.dual = Suffix(direction=Suffix.IMPORT_EXPORT)
        m.ipopt_zL_out = Suffix(direction=Suffix.IMPORT)
        m.ipopt_zU_out = Suffix(direction=Suffix.IMPORT)
        m.ipopt_zL_in = Suffix(direction=Suffix.EXPORT)
        m.ipopt_zU_in = Suffix(direction=Suffix.EXPORT)

        #: K_AUG SUFFIXES  
        m.dcdp = Suffix(direction=Suffix.EXPORT)  #: the dummy constraints
        m.var_order = Suffix(direction=Suffix.EXPORT)  #: Important variables (primal)
        m.rh_name = Suffix(direction=Suffix.IMPORT)
        
        # set which are the variables and which are the parameters for k_aug       
        count_vars = 1
        #print("count_vars:",count_vars)
        if not self._spectra_given:
            pass
        else:
            for t in self.model.times_spectra:
                for c in self._sublist_components:
                    m.C[t, c].set_suffix_value(m.var_order,count_vars)
                    count_vars += 1
        
        if not self._spectra_given:
            pass
        else:
            for l in self.model.meas_lambdas:
                for c in self._sublist_components:
                    m.S[l, c].set_suffix_value(m.var_order,count_vars)
                    count_vars += 1
                        
        for index, value in self.model.Z.items():
            m.Z[index].set_suffix_value(m.var_order, count_vars)
            count_vars += 1
                    
        if self._concentration_given:
            for t in self.model.times_concentration:
                for c in self._sublist_components:
                    m.Z[t, c].set_suffix_value(m.var_order,count_vars)
                    count_vars += 1

        if self._huplc_given:
            raise RuntimeError('Estimability Analysis for additional huplc data is not included as a feature yet!')
                    
        count_dcdp = 1

        idx_to_param = dict()
        for p in m.parameter_names:            
            m.dummyC[p].set_suffix_value(m.dcdp,count_dcdp)
            idx_to_param[count_dcdp]=p
            count_dcdp+=1
          
        #: Clear this file
        with open('ipopt.opt', 'w') as f:
            f.close()
                
        #first solve with Ipopt
        ip = SolverFactory('ipopt')
        solver_results = ip.solve(m, 
                                  symbolic_solver_labels=True,
                                  keepfiles=True,
                                  tee=True,
                                  report_timing=False)

        m.ipopt_zL_in.update(m.ipopt_zL_out)
        m.ipopt_zU_in.update(m.ipopt_zU_out) 
        
        k_aug = SolverFactory('k_aug')
        k_aug.options['dsdp_mode'] = ""  #: sensitivity mode!
        #solve with k_aug in sensitivity mode
        k_aug.solve(m, tee=True)
        print("Done solving sensitivities")

        dsdp = np.loadtxt('dxdp_.dat')
        
        if os.path.exists('dxdp_.dat'):
            os.remove('dxdp_.dat')
        # print(idx_to_param)
        
        return dsdp, idx_to_param

    def rank_params_yao(self, param_scaling=None, meas_scaling=None, sigmas=None):
        """This function ranks parameters in the method described in Yao (2003) by obtaining the 
        sensitivities related to the parameters in the model through solving the original NLP model 
        for concentrations, getting the sensitivities relating to each paramater, and then using 
        them to predict the next sensitivity. User must provide scaling factors as defined in the 
        paper. These are in the form of dictionaries, relating the confidences to the initial
        guesses for the parameters as well as for the confidence in the measurements.

        :param dict param_scaling: Dictionary including each parameter and their relative uncertainty.
          e.g. a value of 0.5 means that the value for the real parameter is within 50% of the guessed value
        :param float meas_scaling: Scalar value showing the certainty of the measurement, obtained from the device
          manufacturer or general knowledge of process
        :param dict sigmasq: Map of component name to noise variance. The map also contains the device noise variance.
        
        :return: list with order of parameters

        """
        if param_scaling is None:
            param_scaling = {}
            print("WARNING: No scaling provided by user, so uncertainties based on the bounds provided by the user is assumed.")
            # uncertainties calculated based on bounds given
            for p in self.model.P:
                lb = self.model.P[p].lb
                ub = self.model.P[p].ub
                init = (ub-lb)/2
                param_scaling[p] = init/(ub-lb)
                print("automated param_scaling", param_scaling)

        elif param_scaling is not None:
            if type(param_scaling) is not dict:
                raise RuntimeError('The param_scaling must be type dict')
        
        if meas_scaling is None:
            meas_scaling = 0.001
            print("WARNING: No scaling for measurments provided by user, so uncertainties based on measurements will be set to 0.01")

        elif meas_scaling is not None:
            if isinstance(meas_scaling, int) or isinstance(meas_scaling, float):
                print("meas_scaling", meas_scaling)
            else:
                raise RuntimeError('The meas_scaling must be type int')
         
        if sigmas is None:
            sigmas = {}
            print("WARNING: No variances provided by user, so variances are assumed to be 1.")
            # sigmas need to be specified
            for p in self.model.P:
                sigmas[p] = 1
                print("automated sigmas", sigmas)
                
        elif sigmas is not None:
            if type(param_scaling) is not dict:
                raise RuntimeError('The param_scaling must be type dict')
            
            else:
                keys = sigmas.keys()
                list_components = [k for k in self._mixture_components]
                all_sigma_specified = True
                for k in list_components:
                    if k not in keys:
                        all_sigma_specified = False
                        sigmas[k] = max(sigmas.values())
                
                if not all_sigma_specified:
                    raise RuntimeError('All variances must be specified to determine sensitivities.\n Please pass '
                                       'variance dictionary to rank_params_yao')
        # k_aug is used to get the sensitivities. The full model is solved with dummy
        # parameters and variables at the initial values for the parameters
        self.cloned_before_k_aug = self.model.clone()
        dsdp, idx_to_param = self.get_sensitivities_for_params(tee=True, sigmasq=sigmas)
        nvars = np.size(dsdp,0)
        nparams = 0
        for v in self.model.P.items():
            if v.is_fixed():
                print(v, end='\t')
                print("is fixed")
                continue
            nparams += 1

        dsdp_scaled = np.zeros_like(dsdp)

        # scale the sensitivities
        i=0
        for k, p in self.model.P.items():
            if p.is_fixed():
                continue            
            for row in range(len(dsdp)):                    
                dsdp_scaled[row][i] = dsdp[row][i]*param_scaling[k]/meas_scaling
            i += 1
        #print("idx_to_param",idx_to_param )
        #print("dsdp_scaled:", dsdp_scaled)
        # euclidean norm for each column of Hessian relating parameters to outputs
        eucnorm = dict()
        eucnorm_scaled = dict()
        count=0
        for i in range(nparams):
            total = 0
            totalscaled=0
            for row in range(len(dsdp)):
                total += dsdp[row][count]**2 
                totalscaled += dsdp_scaled[row][count]**2
            float(total)
            total = np.asscalar(total)            
            sqr = (total)**(0.5)
            eucnorm[count]=sqr
            
            totals = np.asscalar(totalscaled)
            sqrs = (totals)**(0.5)
            eucnorm_scaled[count]=sqrs
            count+=1
        #print("eucnormscaled",eucnorm_scaled)
        
        # sort the norms and link them to the relevant parameters
        sorted_euc = sorted(eucnorm_scaled.values(), reverse=True)

        #print("sorted_euc,", sorted_euc)
        count=0
        ordered_params = dict()
        for p in idx_to_param:
            for t in idx_to_param:
                if sorted_euc[p-1]==eucnorm_scaled[t-1]:
                    ordered_params[count] = t-1
            count +=1
        # print("ordered_params", ordered_params)
        # set the first ranked parameter as the one with highest norm
        iter_count=0
        self.param_ranks[1] = idx_to_param[ordered_params[0]+1]
        # print("self.param_ranks",self.param_ranks)
        #The ranking strategy of Yao, where the X and Z matrices are formed
        next_est = dict()
        X= None
        kcol = None
        countdoub=0
        for i in range(nparams-1):
            if i==0:
                X = np.zeros((nvars,1))
            else:
                X = np.append(X,np.zeros([len(X),1]),1)
            #print(X)
            # Form the appropriate matrix
            for x in range(i+1):
                if x < nparams-countdoub-2:
                    #print(self.param_ranks)
                    # print(x)
                    paramhere = self.param_ranks[(x+1)]
                    #print(paramhere)

                for key, value in self.param_ranks.items():
                    for idx, val in idx_to_param.items():
                        if value ==paramhere:
                            if value == val:
                                #print(key, val, idx)
                                which_col = (idx-1)
                                #print(which_col)
                kcol = dsdp_scaled[:,which_col].T
                recol= kcol.reshape((nvars,1))
                #print("x",x)
                #if x >= 1:
                #    X = np.append(X,np.zeros([len(X),1]),1)
                #    print("why?")
                #    print("X_before 2 loop",X)
                for n in range(nvars):
                    X[n][x] = recol[n][0]
                #print(x)
                #print("X",X)

            #print("X_afterloop",X)
            # Use Ordinary Least Squares to use X to predict Z
            # try is here to catch any error resulting from a singular matrix
            # perhaps not the most elegant way of checking for this
            try:
                A = X.T.dot(X)
                B= np.linalg.inv(A)
                C = B.dot(X.T)
                D=X.dot(C)
                Z = dsdp.T
                Zbar=D.dot(Z.T)
                #Get residuals of prediction
                Res = Z.T - Zbar
            except:
                print("There was an error during the OLS prediction. Most likely caused by a singular matrix. Unable to continue the procedure")
                break

            # Calculate the magnitude of residuals
            magres = dict()
            counter=0
            for i in range(nparams):
                total = 0
                for row in range(len(Res)):
                    total += Res[row][counter]**2
                float(total)
                total = np.asscalar(total)
                sqr = (total)**(0.5)
                magres[counter]=sqr
                counter +=1

            # Sort the residuals and ensure the params are correctly assigned
            sorted_magres = sorted(magres.values(), reverse=True)
            count2=0
            #next_est = dict()
            for p in idx_to_param:
                for t in idx_to_param:
                    if sorted_magres[p-1]==magres[t-1]:
                        # print('p,t', p,t,count2)
                        next_est[count2] = t
                        # print(next_est[count2])
                count2 += 1
            # print(sorted_magres)
            # print("next_est", next_est)
            # Add next most estimable param to the ranking list
            # print('idx_to_param[next_est[0]]', idx_to_param[next_est[0]])
            # print('self.param_ranks[1]',self.param_ranks[1][:])
            if idx_to_param[next_est[0]] not in self.param_ranks.values():
                self.param_ranks[(iter_count+2)]=idx_to_param[next_est[0]]
                iter_count += 1
            else:
                countdoub+=1
            # print("parameter ranks!", self.param_ranks)
            # print("nparams", nparams)

            #print("======================PARAMETER RANKED======================")
            if len(self.param_ranks) == nparams-countdoub-1:
                print("Parameters have been ranked")
                break
        
        #adding the unranked parameters to the list
        #NOTE: if param appears here then it was not evaluated (i.e. it was the least estimable)
        count = 0
        self.unranked_params = {}
        for v,p in self.model.P.items():
            if p.is_fixed():
                print(v, end='\t')
                print("is fixed")
                continue            
            if v in self.param_ranks.values():
                continue
            else:
                self.unranked_params[count]=v
                count += 1

        print("The parameters are ranked in the following order from most estimable to least estimable:")
        count = 0
        for i in self.param_ranks:
            print("Number ", i, "is ", self.param_ranks[i])
            count+=1
        
        print("The least estimable parameters are as follows: ")
        if len(self.unranked_params) == 0:
            print("All parameters ranked")

        for i in self.unranked_params:
            count+=1
            print("unranked ", (count), "is ", self.unranked_params[i])
        
        #preparing final list to return to user
        self.ordered_params = list()
        count = 0
        for i in self.param_ranks:
            self.ordered_params.append(self.param_ranks[i])
            count += 1
        for i in self.unranked_params:
            self.ordered_params.append(self.unranked_params[i])
            count += 1
        print(self.param_ranks)
        return self.ordered_params

    def run_analyzer(self, method=None, parameter_rankings=None, meas_scaling=None, variances=None):
        """This function performs the estimability analysis. The user selects the method to be used. 
        The default will be selected based on the type of data selected. For now, only the method of 
        Wu, McLean, Harris, and McAuley (2011) using the means squared error is used. Other estimability 
        analysis tools will be added in time. The parameter rankings need to be included as well and 
        this can be done using various methods, however for now, only the Yao (2003) method is used.

        :param str method: The estimability method to be used. Default is Wu, et al (2011) for concentrations. Others
             to be added
        :param list parameter_rankings: A list containing the parameter rankings in order from most estimable to least
             estimable. Can be obtained using one of Kipet's parameter ranking functions.
        :param float meas_scaling: Value showing the certainty of the measurement obtained from the device manufacturer
             or general knowledge of process. Same as used in the parameter ranking algorithm.
        :param dict variances: Variances are required, as needed by the parameter estimator.
        
        :return: List of parameters that should remain in the parameter estimation, while all other  parameters
            should be fixed.

        """
        if method is None:
            method = "Wu"
            print("The method to be used is that of Wu, et al. 2011")
        elif method != "Wu":
            print("The only supported method for estimability analysis is that of Wu, et al., 2011, currently")
        else:
            method = "Wu"
            
        if parameter_rankings is None:
            raise RuntimeError('The parameter rankings need to be provided in order to run the estimability analysis chosen')
            
        elif parameter_rankings is not None:
            if type(parameter_rankings) is not list:
                raise RuntimeError('The parameter_rankings must be type dict')   
                
        for v,k in self.model.P.items(): 
            if v in parameter_rankings:
                continue
            else:
                print("Warning, %s is not included in the parameter rankings algorithm" % v)
                
        for v in parameter_rankings:
            if v not in self.model.P:
                raise RuntimeError("parameter %s is not in the model! Either remove the parameter from the list or add it to the model" % v)
        
        if meas_scaling is None:
            meas_scaling = 0.001
            print("WARNING: No scaling for measurments provided by user, so uncertainties based on measurements will be set to 0.01")
        elif meas_scaling is not None:
            if isinstance(meas_scaling, int) or isinstance(meas_scaling, float):
                pass
            else:
                raise RuntimeError('The meas_scaling must be type int')
                
        if variances is None:
            variances ={}
            print("WARNING: No variances provided by user, so variances are assumed to be 1.")
            # sigmas need to be specified
            for p in self.model.P:
                variances[p] = 1
                print("automated sigmas", variances)
            variances["device"] = 1
        elif variances is not None:
            if type(variances) is not dict:
                raise RuntimeError('The sigmas must be type dict')
        
        if method == "Wu":
            estimable_params = self.wu_estimability(parameter_rankings, meas_scaling, variances)
            return estimable_params
        else:
            raise RuntimeError("the estimability method must be 'Wu' as this is the only supported method as of now")


    def wu_estimability(self, parameter_rankings = None, meas_scaling = None, sigmas = None):
        """This function performs the estimability analysis of Wu, McLean, Harris, and McAuley (2011) 
        using the means squared error. 

        :param list parameter_rankings: A list containing the parameter rankings in order from most estimable to least
           estimable. Can be obtained using one of Kipet's parameter ranking functions.
        :param int meas_scaling: Measurement scaling as used to scale the sensitivity matrix during param ranking
        :param dict sigmas: Dictionary containing all the variances as required by the parameter estimator
        
        :return: List of parameters that should remain in the parameter estimation, while all other parameters should be fixed.
        :rtype: list

        """
        J = dict()
        params_estimated = list()
        cloned_full_model = dict()
        cloned_pestim = dict()
        results = dict()
        # For now, instead of using Levenberg-Marquardt least squares, we will use Kipet to perform the estimation
        # of every model. Here we generate each of the simplified models.
        # first we clone the main model so that we work with the full model at every iteration
        count = 1
        for p in parameter_rankings:
            cloned_full_model[count] = self.cloned_before_k_aug.clone()
            count += 1
        count = 1
        # Then we go create each simplified model, fixing remaining variables
        for p in parameter_rankings:
            params_estimated.append(p)            
            #print("performing parameter estimation for: ", params_estimated)
            for v,k in cloned_full_model[count].P.items():
                if v in params_estimated:
                    continue
                else:
                    #print("fixing the parameters for:",v,k)
                    #fix parameters not in simplified model
                    ub = value(cloned_full_model[count].P[v])
                    lb = ub
                    cloned_full_model[count].P[v].setlb(lb)
                    cloned_full_model[count].P[v].setub(ub)
            # We then solve the Parameter estimaion problem for the SM
            options = dict()            
            cloned_pestim[count] = ParameterEstimator(cloned_full_model[count])
            if count>=2:
                if hasattr(results[count-1], 'Y'):
                    cloned_pestim[count].initialize_from_trajectory('Y', results[count - 1].Y)
                    cloned_pestim[count].scale_variables_from_trajectory('Y', results[count - 1].Y)
                if hasattr(results[count-1], 'X'):
                    cloned_pestim[count].initialize_from_trajectory('X', results[count - 1].X)
                    cloned_pestim[count].scale_variables_from_trajectory('X', results[count - 1].X)
                if hasattr(results[count-1], 'C'):
                    cloned_pestim[count].initialize_from_trajectory('C', results[count - 1].C)
                    cloned_pestim[count].scale_variables_from_trajectory('C', results[count - 1].C)
                # if hasattr(results[count-1], 'S'):
                #     cloned_pestim[count].initialize_from_trajectory('S', results[count-1].S)
                #     cloned_pestim[count].scale_variables_from_trajectory('S', results[count-1].S)
                cloned_pestim[count].initialize_from_trajectory('Z', results[count-1].Z)
                cloned_pestim[count].scale_variables_from_trajectory('Z', results[count - 1].Z)
                cloned_pestim[count].initialize_from_trajectory('dZdt', results[count - 1].dZdt)
                cloned_pestim[count].scale_variables_from_trajectory('dZdt', results[count - 1].dZdt)

            results[count] = cloned_pestim[count].run_opt('ipopt',
                                        tee=True,#False,
                                        solver_opts = options,
                                        variances=sigmas, symbolic_solver_labels=True
                                        )

            # print('TC',TerminationCondition.optimal)
            # print('selfterm',self.termination_condition)

            for v,k in results[count].P.items():
                print(v,k)
            # Then compute the scaled residuals to obtain the Jk in the Wu et al paper   
            J [count] = self._compute_scaled_residuals(results[count], meas_scaling)
            count += 1            
        #print(J)
        count = count - 1
        # Since the estimability procedure suggested by Wu will always skip the last
        # parameter, we should check whether all parameters can be estimated
        # For now this is done by checking that the final parameter does not provide a massive decrease
        # the residuals
        low_MSE = J[1]
        #print("J",J)
        #print("low_MSE", low_MSE)
        listMSE = list()
        for k in J:
            listMSE.append(J[k]) 
            if J[k] <= low_MSE:
                low_MSE = J[k]
            else:
                continue
        #print(count)
        #print(J[count])
        #print(low_MSE)
        if J[count] == low_MSE:
            print("Lowest MSE is given by the lowest ranked parameter, therefore the full model should suffice")
            listMSE.sort()
            print("list of ordered mean squared errors of each :")
            print(listMSE)
            if listMSE[0]*10 <= listMSE[1]:
                print("all parameters are estimable! No need to reduce the model")
                return parameter_rankings

        # Now we move the the final steps of the algorithm where we compute critical ratio
        # and the corrected critical ratio
        # first we need the total number of responses
        N = 0
        for c in self._sublist_components:
            for t in self.model.allmeas_times:
                N += 1 
                
        crit_rat = dict()
        cor_crit_rat = dict()
        for k in J:
            if k == count:
                break
            crit_rat[k] = (J[k] - J[count])/(count - k)
            crit_rat_Kub = max(crit_rat[k]-1,crit_rat[k]*(2/(count - k + 2)))
            cor_crit_rat[k] = (count - k)/N * (crit_rat_Kub - 1)
        
        #Finally we select the value of k with the lowest corrected critical value
        params_to_select = min(cor_crit_rat, key=lambda x: cor_crit_rat.get(x) )
        print("The number of estimable parameters is:", params_to_select)
        print("optimization should be run wih the following parameters as variables and all others fixed")
        estimable_params = list()
        count=1
        for p in parameter_rankings:
            print(p)
            estimable_params.append(p)
            if count >= params_to_select:
                break
            count += 1
        return estimable_params

    def _compute_scaled_residuals(self, model, meas_scaling = None):
        """Computes the square of residuals between the optimal solution (Z) and the concentration data (C)
        
        :param ConcreteModel model: Solved pyomo model results object
        :param dict meas_scaling: parameter scaling, defined in Wu, needs to be the same as used to rank
            parameters (scale sensitivity matrix)

        :return: Value of sum of squared scaled residuals

        """
        self.residuals = dict()
        
        E = 0
        for c in self._sublist_components:
            for t in self.model.allmeas_times:
                a = model.Cm[c][t]
                b = model.Z[c][t]
                r = ((a - b) ** 2)
                self.residuals[t, c] = r
                   
        
        for c in self._sublist_components:
            for t in self.model.allmeas_times:
                E += self.residuals[t, c] / (meas_scaling ** 2)

        # print(E)
        # print(self.residuals)
        # import pandas as pd


        # residuals = pd.DataFrame(
        #     np.zeros((len(self.model.allmeas_times), len(self._sublist_components))),
        #     columns=self._sublist_components,
        #     index=self.model.allmeas_times
        #     )
        
        # for index, values in self.model.Cm.items():
        #     residuals.loc[index] = (model.Z[index].value - model.Cm[index].value) ** 2 / meas_scaling ** 2

        # print(index)
        # print(residuals.loc[index])

        # print(residuals.tail())
        # E2 = sum(sum(residuals.values))
        # print(E2)
        return E






__version__ = "1.0.0"
import kipet
from kipet.general_settings.settings import Settings
from kipet.main_modules.reaction_set import ReactionSet
from kipet.main_modules.reaction_model import ReactionModel
from kipet.input_output.kipet_io import *
kipet.settings = Settings()



"""
Helper functions for making KIPET models
"""
from kipet.general_settings.variable_names import VariableNames

__var = VariableNames()


def step_fun(model, time_var, num='0', coeff=1, time=1e-3, fixed=True, switch=False, M=1, eta=1e-2, constant=0.5, bounds=None):
    """This formulates the step functions for KIPET using a sigmoidal function

    :param ConcreteModel model: The ConcreteModel using the step function
    :param float time_var: Time point for step
    :param str num: The name of the step function
    :param float coeff: Coefficient for the step size
    :param float time: The time for the step change (init if not fixed)
    :param bool fixed: Choose if the time is known and fixed or variable
    :param bool/str switch: True if turning on, False if turning off, optional string args too
    :param float M: Sigmoidal tuning parameter
    :param float eta: Tuning parameter
    :param float constant: Constant added to bring to certain value
    :param tuple bounds: Bounds for the time if not fixed

    :return: Returns the pyomo expression for the step function
    :rtype: expression

    """
    if switch in ['on', 'turn_on']:
        switch = True
    elif switch in ['off', 'turn_off']:
        switch = False
    
    factor = 1 if not switch else -1
    
    if fixed:
        if time is None:
            raise ValueError('Time cannot be None')
        time_var_step = factor*time
    else: 
        if time is not None:
            model.time_step_change[num].set_value(time)
        time_var_step = factor*model.time_step_change[num]
        
    pyomo_var = getattr(model, __var.dosing_variable)[time_var, __var.dosing_component]
    step_point = coeff*(0.5*((time_var_step - factor*pyomo_var) / ((time_var_step - factor*pyomo_var)**2 + eta**2/M) ** 0.5) + constant)
    
    return step_point



"""
This module contains functions for ensuring that variables have congruent time indicies.

For many users who may have delved into older versions of KIPET, these were formerly found in the PyomoSimulator and
Simulator classes.
"""
# Standard library imports
import bisect

# Third party imports
import numpy as np
import pandas as pd


def interpolate_trajectory(t, tr):
    """Method for interpolating trajectories

    :param list t: the list of times
    :param pandas.DataFrame tr: The trajectory data

    :return float tr_val: The interpolated value

    """
    tr_val = np.zeros((len(t)))
    times = [float(ti) for ti in t]
    indx = tr.index.astype(float)

    for i, t_indx in enumerate(times):

        if i == 0:
            tr_val[i] = tr.iloc[0]
            continue

        indx_left = bisect.bisect_left(indx[1:], times[i])
        if indx_left == len(tr) - 1:
            tr_val[i] = tr.iloc[indx_left]
            continue

        dx = indx[indx_left + 1] - indx[indx_left]
        dy = tr.iloc[indx_left + 1] - tr.iloc[indx_left]
        m = dy / dx
        val = tr.iloc[indx_left] + (times[i] - indx[indx_left]) * m
        tr_val[i] = val

    return tr_val


def interpolate_trajectory2(t, data):
    """Takes some data and fills in the missing points using interpolation

    :param float t: The time point
    :param pandas.DataFrame data: The data to interpolate with

    :return df_interpolated: The data after interpolation
    :rtype: pandas.DataFrame

    """
    data = pd.DataFrame(data)
    df_interpolated = pd.DataFrame(columns=data.columns)
    
    for col in data.columns:
        
        tr = data[col]
        tr_val = np.zeros((len(t)))
        times = [float(ti) for ti in t]
        indx = tr.index.astype(float)
    
        for i, t_indx in enumerate(times):
    
            if i == 0:
                tr_val[i] = tr.iloc[0]
                continue        
    
            indx_left = bisect.bisect_left(indx[1:], times[i])
            if indx_left == len(tr) - 1:
                tr_val[i] = tr.iloc[indx_left]
                continue
            
            dx = indx[indx_left + 1] - indx[indx_left]
            dy = tr.iloc[indx_left + 1] - tr.iloc[indx_left]
            m = dy/dx
            val = tr.iloc[indx_left] +  (times[i] - indx[indx_left]) * m
            tr_val[i] = val
    
        df_interpolated[col] = tr_val
    
    return df_interpolated


"""
Component expression handling and units model development
"""
# KIPET library imports
from kipet.model_tools.pyomo_model_tools import get_index_sets
from kipet.general_settings.variable_names import VariableNames

string_replace_dict = {
    ' ': '_',
    '-': 'n',
    '+': 'p',
    '.': '_',
    }


class ModalVar:
    """Simple class to model a model variable"""
    def __init__(self, name, comp, index, model):
        
        self.name = name
        self.comp = comp
        self.index = index
        self.model = model


class Comp:
    """This class is needed to handle pyomo variables and return a dict of these variables.

    This will not be used by the user directly. This is one of the key classes that allows the user to use
    dummy pyomo variables in building the KIPET models.

    :param ConcreteModel model: The model we are going to use (dummy model)
    :param
    """
    
    var = VariableNames()
    
    def __init__(self, model):
        
        self._model = model
        self._model_vars = self.var.model_vars
        self._rate_vars = self.var.rate_vars
        self.var_dict = {}
        self.var_unit_dict = {}
        self.assign_vars()
        self.assign_rate_vars()
        
    def assign_vars(self):
        """Digs through and assigns the variables as top-level attributes

        :return: None
        """
        list_of_vars = self._model_vars
        
        for mv in list_of_vars:
            if hasattr(self._model, mv):
                mv_obj = getattr(self._model, mv)
                index_sets = get_index_sets(mv_obj)
                
                dim = len(index_sets)
                if dim > 1:
                    indx = 1
                else:
                    indx = 0
                
                comp_set = list(index_sets[indx].keys())
                for comp in comp_set:
                    
                    comp = str(comp)
                    if isinstance(comp, str):
                        if isinstance(comp[0], int):
                            comp.insert(0, 'y')
                        
                        comp_name = comp
                        for k, v in string_replace_dict.items():
                            comp_name.replace(k, v)
                    
                    self.var_dict[comp_name] = ModalVar(comp_name, comp_name, mv, self._model)
                    
                    if dim > 1:
                        setattr(self, comp_name, mv_obj[0, comp])
                    else:
                        setattr(self, comp_name, mv_obj[comp])

        return None

    def assign_rate_vars(self):
        """Assigns the rate variables as top-level attributes

        :return: None

        """
        for mv in self._rate_vars:
            if hasattr(self._model, mv):
                
                mv_obj = getattr(self._model, mv)
                index_sets = get_index_sets(mv_obj)
                comp_set = list(index_sets[-1].keys())
                dim = len(index_sets)
                
                for comp in comp_set:
                    
                    if isinstance(comp, str):
                        comp_name = comp
                        comp_name.replace(' ', '_')
                    elif isinstance(comp, int):
                        comp_name = f'y{comp}'
                    
                    comp_name = f'd{comp_name}dt'
                    self.var_dict[comp_name] = ModalVar(comp_name, str(comp), mv, self._model)
                    if dim > 1:
                        setattr(self, comp_name, mv_obj[0, comp])
                    else:
                        setattr(self, comp_name, mv_obj[comp])

        return None
    
    @property
    def get_var_list(self):
        """Returns a list of variables in the Comp dict

        :return: A list of variables
        :rtype: list

        """
        return list(self.var_dict.keys())



"""
Problem Generation tools for Kipet
"""
import numpy as np
import pandas as pd


def gaussian_single_peak(wl, alpha, beta, gamma):
    """
    Helper function to generate absorption data based on 
    Lorentzian parameters
    
    :param float wl: Wavelength
    :param float beta: Beta
    :param float alpha: Alpha
    :param float gamma: gamma
    
    :return: The single peak
    :rtype: float

    """
    return alpha*np.exp(-(wl-beta)**2/gamma)


def absorbance(wl, alphas, betas, gammas):
    """
    Helper function to generate absorption data based on 
    Lorentzian parameters
    
    :param float wl: Wavelength
    :param array-like betas: Beta
    :param array-like alphas: Alpha
    :param array-like gammas: gamma
    
    :return: The sum of single peaks
    :rtype: float

    """
    return sum(gaussian_single_peak(wl,alphas[i],betas[i],gammas[i]) for i in range(len(alphas)))


def generate_absorbance_data(wl_span, parameters_dict):
    """
    Helper function to generate absorption data based on 
    Lorentzian parameters
    
    :param array-like wl_span: Array of wavelengths
    :param dict parameters_dict: The dictionary of alphas, betas, and gammas
   
    :return: data_frame
    :rtype: pandas.DataFrame

    """
    components = parameters_dict.keys()
    n_components = len(components)
    n_lambdas = len(wl_span)
    array = np.zeros((n_lambdas,n_components))
    for i,l in enumerate(wl_span):
        j = 0
        for k, p in parameters_dict.items():
            alphas = p['alphas']
            betas  = p['betas']
            gammas = p['gammas']
            array[i,j] = absorbance(l,alphas,betas,gammas)
            j+=1

    data_frame = pd.DataFrame(data=array,
                              columns = components,
                              index=wl_span)
    return data_frame


def generate_random_absorbance_data(wl_span, component_peaks, 
                                    component_widths=None, seed=None):
    """
    Helper function to generate absorption data based on 
    Lorentzian parameters
    
    :param array-like wl_span: Array of wavelengths
    :param dict component_peaks: Dictionary with number of component peaks
    :param dict component_widths: Optional widths for components (otherwise 1000)
    :param int seed: Random seed number
   
    :return: data_frame
    :rtype: pandas.DataFrame

    """
    np.random.seed(seed)
    parameters_dict = dict()
    min_l = min(wl_span)
    max_l = max(wl_span)
    #mean=1000.0
    #sigma=1.5*mean
    for k,n_peaks in component_peaks.items():
        params = dict()
        if component_widths:
            width = component_widths[k]
        else:
            width = 1000.0
        params['alphas'] = np.random.uniform(0.1,1.0,n_peaks)
        params['betas'] = np.random.uniform(min_l,max_l,n_peaks)
        params['gammas'] = np.random.uniform(1.0,width,n_peaks)
        parameters_dict[k] = params

    return generate_absorbance_data(wl_span,parameters_dict)


def add_noise_to_signal(signal, size):
    """
    Adds a random normally distributed noise to a clean signal. Used mostly in Kipet
    To noise absorbances or concentration profiles obtained from simulations. All
    values that are negative after the noise is added are set to zero
    
    :param pandas.DataFrame signal: The Z or S matrix to have noise added to it
    :param float size: sigma (or size of distribution)
    Returns:
    :return: Noised data
    :rtype: pandas.DataFrame
    
    """
    clean_sig = signal    
    noise = np.random.normal(0,size,clean_sig.shape)
    sig = clean_sig+noise    
    df= pd.DataFrame(data=sig)
    df[df < 0] = 0
    return df



"""
General KIPET Model Varialbe names
"""

class VariableNames(object):
    
    """Names of the KIPET model variables

    This provides a central location to reduce errors in developing KIPET modules by forcing a naming convention.
    It also provides useful collections of variables for use in many other functions.

    This is not really needed by the user and should only be used by developers.
    """
    def __init__(self):
    
        # Strong defaults
        self.volume_name = 'V'
    
        # Variables for KIPET models    
        self.model_parameter = 'P'
        self.model_parameter_scaled = 'K'
        self.concentration_measured = 'Cm'
        self.concentration_spectra = 'C'
        #self.concentration_spectra_abs = 'Cs'
        self.concentration_model = 'Z'
        self.concentration_model_rate = f'd{self.concentration_model}dt'
        self.state = 'U'
        self.state_model = 'X'
        self.state_model_rate = f'd{self.state_model}dt'
        self.spectra_species = 'S'
        self.spectra_data = 'D'
        self.user_defined = 'UD'
        self.huplc_data = 'Dhat'
        self.smooth_parameter = 'Ps'
        self.algebraic = 'Y'
        self.unwanted_contribution = 'g'
        self.ode_constraints = 'odes'
        
        self.dosing_variable = 'Dose'
        self.dosing_component = 'd_var'
        
        self.time_step_change = 'time_step_change'
        self.step_variable = 'step'
        # Debug options
        
        self.model_constant = 'Const'
        self.concentration_init = 'Pinit'
        
        self.DEBUG = False
        
        
    @property
    def optimization_variables(self):
        """These are the independent variables that need to be fixed in 
        simulations
        """
        
        model_vars = [self.model_parameter,
                      self.time_step_change,
                      self.concentration_init
                      ]
        
        return model_vars
    
    @property
    def time_dependent_variables(self):
        """Property to return variables that are time dependent:

        :Defaults:

            ['Z', 'dZdt', 'S', 'C', 'X', 'dXdt', 'U', 'Y']

        :return model_vars: The list of target variables
        :rtype: list

        """
        model_vars = [self.concentration_model,
                      self.concentration_model_rate,
                      self.spectra_species,
                      self.concentration_spectra,
                      self.state_model,
                      self.state_model_rate,
                      self.state,
                      self.algebraic,
                      ]
        
        return model_vars
        
    @property
    def modeled_states(self):
        """Property to return variables that are modeled:

        :Defaults:

            ['Z', 'dZdt', 'X', 'dXdt']

        :return model_vars: The list of target variables
        :rtype: list

        """
        model_vars = [self.concentration_model,
                      self.concentration_model_rate,
                      self.state_model,
                      self.state_model_rate,
                      ]
        
        return model_vars
    
    @property
    def model_vars(self):
        """Property to return component variables:

        :Defaults:

            ['Z', 'X', 'P', 'Y', 'step', 'Const']

        :return model_vars: The list of target variables
        :rtype: list

        """
        model_vars = [self.concentration_model,
                      self.state_model,
                      self.model_parameter,
                      self.algebraic,
                      self.step_variable,
                      self.model_constant,
                      ]
        
        return model_vars
    
    @property
    def rate_vars(self):
        """Property to return rate variables:

        :return model_vars: The list of target variables
        :rtype: list

        """
        model_vars = [
            self.concentration_model_rate,
            self.state_model_rate,
            ]
        
        return model_vars
    
    @property
    def plot_vars(self):
        """Property to return plotted variables

        :return model_vars: The list of target variables
        :rtype: list

        """
        model_vars = [
            self.concentration_model_rate,
            self.state_model_rate,
            self.algebraic,
            self.step_variable,
            self.spectra_species,
            ]
        
        return model_vars



"""TemplateBuilder - handles inputs for generating the Pyomo model"""

# Standard library imports
import inspect
import itertools
import logging
import warnings

# Third party imports
import numpy as np
import pandas as pd
import six
from pyomo.dae import ContinuousSet, DerivativeVar
from pyomo.environ import (ConcreteModel, Constraint, ConstraintList, 
                           Param, Reals, Var, Set, Suffix)

# KIPET library imports
from kipet.model_components.component_expression import Comp
from kipet.model_tools.visitor_classes import ReplacementVisitor
from kipet.general_settings.variable_names import VariableNames

logger = logging.getLogger('ModelBuilderLogger')


class TemplateBuilder(object):
    """Helper class for creation of models.
    
    This class takes the arguments, settings, and data from the ReactionModel
    and uses it to generate the Pyomo model used in simulation or parameter
    estimation. The TemplateBuilder object is not meant to be accessed directly
    by the user.

    """
    __var = VariableNames()

    def __init__(self, **kwargs):
        """Initialization for the TemplateBuilder object.

        """
        self._smoothparameters = dict()  # added for mutable parameters CS
        self._smoothparameters_mutable = dict() #added for mutable parameters CS

        #added for initial condition parameter estimates: CS
        self._initextraparams = dict()
        self._initextraparams_init = dict()  # added for parameter initial guess CS
        self._initextraparams_bounds = dict()
        self._y_bounds = dict()  # added for additional optional bounds CS
        self._y_init = dict()
        self._prof_bounds = list()  # added for additional optional bounds MS
        self._prof_points = list()
        self._point_times = list()
        self._point_wavelengths = list()
        self._spectral_data = None
        self._concentration_data = None
        self._complementary_states_data = None # added for complementary state data (Est.) KM
        self._custom_data = None
        self._custom_objective = None
        self._huplc_data = None #added for additional data CS
        self._smoothparam_data = None  # added for additional smoothing parameter data CS
        self._absorption_data = None
        self._odes = None
        self._meas_times = set()
        self._huplcmeas_times = set() #added for additional data CS
        self._allmeas_times = set() #added for new data structure CS
        self._m_lambdas = None
        self._complementary_states = set()
        self._algebraics = dict()
        self._algebraic_constraints = None
        self._known_absorbance = None
        self._is_known_abs_set = False
        self._warmstart = False  # add for warmstart CS
        self._known_absorbance_data = None
        self._non_absorbing = None
        self._is_non_abs_set = False
        self._huplc_absorbing = None #for add additional data CS
        self._is_huplc_abs_set = False #for add additional data CS
        self._vol = None #For add additional data CS
        self._feed_times = set()  # For inclusion of discrete feeds CS
        self._is_D_deriv = False
        self._is_C_deriv = False
        self._is_U_deriv = False
        self._is_Dhat_deriv = False
        self._state_sigmas = None # Need to put sigmas into the pyomo model as params
        self._model_constants = None # Used in EstimaationPotential
        self._scale_parameters = False # Should be True for EstimationPotential (automatic)
        self._times = None
        self._all_state_data = list()
        self._init_absorption_data = None
        self._G_contribution = None
        self._qr_bounds = None
        self._qr_init = None
        self._g_bounds = None
        self._g_init = None
        self._add_dosing_var = False
        self._solid_spec = None

    def add_state_variance(self, sigma_dict):
        """Provide a variance for the measured states
        
        :param dict sigma_dict: A dictionary of the measured states with known variance. Provide the value of
            sigma (standard deviation).
            
        :return: None

        """
        self._state_sigmas = sigma_dict 
        
        return None

    def set_parameter_scaling(self, use_scaling):
        """Makes an option to use the scaling method implemented for estimability

        :param bool use_scaling: Defaults to False, for using scaled parameters
            
        :return: None
        
        """
        self._scale_parameters = use_scaling
    
        return None
        
    def set_model_times(self, times):
        """Add the times to the builder template.
        
        :param tuple times: start and end times for the pyomo model
        
        :return: None
            
        """
        self._times = times

    def add_model_element(self, BlockObject):
        """Adds the ReactionModel component blocks

        :param ComponentBlock BlockObject: The component block from ReactionModel

        :return None:

        """
        data_type = BlockObject.attr_class_set_name.rstrip('s')
        block_attr_name = f'template_{data_type}_data'
        setattr(self, block_attr_name, BlockObject)
        return None
 
    def add_smoothparameter(self, *args, **kwds):
        """Add kinetic parameter(s) to the model.

        :param tuple args: name (str) and dataframe (pandas.DataFrame)
        :param dict kwds: Arguments, mutable as bool

        :return: None

        """
        mutable = kwds.pop('mutable', False)

        if len(args) == 2:
            first = args[0]
            second = args[1]
            if isinstance(first, six.string_types):
                self._smoothparameters[first] = second
                if mutable is not False and second is pd.DataFrame:
                    self._smoothparameters_mutable[first] = mutable #added for mutable parameters CS
            else:
                raise RuntimeError('Parameter argument not supported. Try pandas.Dataframe and mutable=True')
        else:
            raise RuntimeError('Parameter argument not supported. Try pandas.Dataframe and mutable=True')
    
    def input_data(self, data_block_dict=None, spectral_data=None):
        """Inputs the data from the ReactionModel. This will also figure out the times for the models if not provided.

        :param dict data_block_dict: The input data
        :param bool spectral_data: Indicates if the data is spectral

        :return: None
        
        """
        time_span = 0
        time_conversion_factor = 1
        
        data_type_labels = {'component' : 'concentration',
                            'state': 'complementary_states',
                            'algebraic': 'custom',
                            'uplc': ''
                            }

        if data_block_dict is not None:
            for data_type, data_label in data_type_labels.items():
                if hasattr(self, f'template_{data_type}_data'):
                    c_info = getattr(self, f'template_{data_type}_data')
                    for comp in c_info:
                        if hasattr(comp, 'data_link'):
                            data_block = data_block_dict[comp.data_link]
                            time_span = max(time_span, data_block.time_span[1]*time_conversion_factor)
                            data_frame = data_block.data[comp.name]
                            if not comp.use_orig_units:
                                data_frame *= comp.conversion_factor
                            # data_frame.index = data_frame.index*time_conversion_factor
                            self._add_state_data(data_frame, data_label, overwrite=False)

        # Spectral data is handled differently
        if spectral_data is not None:
            self._add_state_data(spectral_data.data, 'spectral')
            # spectral_data.data.index = spectral_data.data.index*time_conversion_factor
            time_span = max(time_span, spectral_data.data.index.max())
                
        self.time_span_max = time_span
       
        return None
    
    def clear_data(self):
        """Method to clear the model data

        :return: None

        """
        self._spectral_data = None
        self._concentration_data = None
        self._complementary_states_data = None # added for complementary state data (Est.) KM
        self._huplc_data = None #added for additional data CS
        self._smoothparam_data = None  # added for additional smoothing parameter data CS
        self._absorption_data = None
        
        return None

    def _add_state_data(self, data, data_type, label=None, overwrite=True):
        """Generic method for adding data (concentration or complementary 
        state data) - uses the measured data attribute to process
        
        :param pandas.DataFrame data: DataFrame with measurement times as indices and concentrations as columns.
        :param str data_type: The name of the attribute for where the data is to be stored.
        :param str label: The label used to descibe the data in the pyomo model index.

        :return: None

        """
        built_in_data_types = {
            'concentration' : self.__var.concentration_measured,
            'complementary_states' : self.__var.state,
            'spectral' : self.__var.spectra_data,
            'huplc' : self.__var.huplc_data,
            'smoothparam' : self.__var.smooth_parameter,
            'custom' : self.__var.user_defined,
            }
        
        state_data = [
            self.__var.concentration_spectra,
            self.__var.state,
            self.__var.concentration_measured,
            self.__var.user_defined,
            self.__var.huplc_data,
            ]
            
        deriv_data = [
            self.__var.concentration_spectra,
            self.__var.state,
            self.__var.spectra_data,
            self.__var.huplc_data,
            self.__var.concentration_measured,
            self.__var.user_defined,
            ]
        
        if label is None:
            try:
                label = built_in_data_types[data_type]
            except:
                raise ValueError("You need to provide a label for custom data types")
        
        if isinstance(data, pd.Series):
            data = pd.DataFrame(data)
        
        if isinstance(data, pd.DataFrame):
            dfc = pd.DataFrame(index=self._feed_times, columns=data.columns)
            for t in self._feed_times:
                if t not in data.index:
                    dfc.loc[t] = [0.0 for n in range(len(data.columns))]
      
            dfallc = data.append(dfc)
            dfallc.sort_index(inplace=True)
            dfallc.index = dfallc.index.to_series().apply(
                lambda x: np.round(x, 6))

            count = 0
            for j in dfallc.index:
                if count >= 1 and count < len(dfallc.index):
                    if dfallc.index[count] == dfallc.index[count - 1]:
                        dfallc = dfallc.dropna()
                        
                count += 1

            dfallc = dfallc.dropna(how='all')

            if not overwrite:
                if hasattr(self, f'_{data_type}_data'):
                    df_data = getattr(self, f'_{data_type}_data')
                    df_data = pd.concat([df_data, dfallc], axis=1)
                    setattr(self, f'_{data_type}_data', df_data)
            else:
                setattr(self, f'_{data_type}_data', dfallc)
            
            if label in state_data:
                self._all_state_data += list(data.columns)
        else:
            raise RuntimeError(f'{data_type.capitalize} data format not supported. Try pandas.DataFrame')
        
        if label in deriv_data:
            C = np.array(dfallc)
            setattr(self, f'_is_{label}_deriv', False)
            for t in range(len(dfallc.index)):
                for l in range(len(dfallc.columns)):
                    if C[t, l] >= 0 or np.isnan(C[t, l]):
                        pass
                    else:
                        setattr(self, f'_is_{label}_deriv', True)
            if getattr(self, f'_is_{label}_deriv') == True:
                print(
                    f"Warning! Since {label}-matrix contains negative values Kipet is assuming a derivative of {label} has been inputted")

        return None
           
    def add_huplc_data(self, data, overwrite=True):
        """Add HPLC or UPLC data as a wrapper to _add_state_data

        :param pandas.DataFrame data: DataFrame with measurement times as indices and wavelengths as columns.
        :param bool overwrite: Option to overwrite exisiting data

        :return: None

        """
        self._add_state_data(data,
                             data_type='huplc',
                             overwrite=overwrite)
        
        self._is_huplc_abs_set = True
        self._huplc_absorbing = data.columns

    def add_smoothparam_data(self, data, overwrite=True):
        """Add smoothing parameter data

        :param pandas.DataFrame data: DataFrame with measurement times as indices and wavelengths as columns.
        :param bool overwrite: Option to overwrite exisiting data

        :return: None

        """
        self._add_state_data(data,
                             data_type='smoothparam',
                             overwrite=overwrite)

    def add_init_absorption_data(self, data):
        """Initialize the absorption using simulated data

        :param pandas.DataFrame data: DataFrame with measurement times as indices and wavelengths as columns.

        :return: None

        """
        if isinstance(data, pd.DataFrame):
            self._init_absorption_data = data
        else:
            raise RuntimeError('Spectral data format not supported. Try pandas.DataFrame')

    def add_absorption_data(self, data, overwrite=True):
        """Add absorption data

        :param pandas.DataFrame data: DataFrame with measurement times as indices and wavelengths as columns.
        :param bool overwrite: Option to overwrite exisiting data

        :return: None

        """
        if isinstance(data, pd.DataFrame):
            self._absorption_data = data
        else:
            raise RuntimeError('Spectral data format not supported. Try pandas.DataFrame')

    @staticmethod
    def round_time(time):
        """Round the time to a fixed number of significant digits

        :param float time: The time value

        :return: time rounded to fixed number of decimals (6)
        :rtype: float

        """
        return round(time, 6)

    def add_feed_times(self, times):
        """Add measurement times to the model

        :param array-like times: feeding points

        :return: None

        """
        for t in times:
            t = self.round_time(t)
            self._feed_times.add(t)

    def add_qr_bounds_init(self, bounds=None, init=None):
        """Read bounds and initialize for qr variables

        :param tuple bounds: The qr bounds
        :param float init: The initial qr value

        :return: None

        """
        self._qr_bounds = bounds
        self._qr_init = init
        
        return None
        
    def add_g_bounds_init(self, bounds=None, init=None):
        """Read bounds and initialize for g variables

        :param tuple bounds: The qr bounds
        :param float init: The initial qr value

        :return: None

        """
        self._g_bounds = bounds
        self._g_init = init
    
        return None
    
    def add_dosing_var(self, n_steps):
        """Adds a dosing variable to the template

        :param int n_steps: Number of steps

        :return: None
        """
        self._add_dosing_var = True
        self._number_of_steps = n_steps
    
    def set_odes_rule(self, rule):
        """Defines the ordinary differential equations that define the dynamics of the model

        :param dict rule: Model expressions for the rate equations

        :return: None

        """
        self._odes = rule
        
        return None

    def set_algebraics_rule(self, rule, asdict=False):
        """Defines the algebraic equations for the system

        :param dict rule: Dictionary of algebraic expressions

        :return: None

        """
        if not asdict:
            inspector = inspect.getargspec(rule)
            if len(inspector.args) != 2:
                raise RuntimeError('The rule should have two inputs')
            self._algebraic_constraints = rule
        else:
            self._algebraic_constraints = rule
            self._use_alg_dict = True
        
        return None

    def set_objective_rule(self, algebraic_vars):
        """Set the algebraic expressions.

        Defines the algebraic equations for the system

        :param list algebraic_vars: List of algebraic variables

        :return: None

        """
        if not isinstance(algebraic_vars, list):
            algebraic_vars = list(algebraic_vars)
        self._custom_objective = algebraic_vars
        
        return None

    def bound_profile(self, var, bounds, comp=None, profile_range=None):
        """function that allows the user to bound a certain profile to some value

        :param GeneralVar var: The pyomo variable that we will bound
        :param str comp:The component that bound applies to
        :param tuple profile_range: The range within the set to be bounded
        :param tuple bounds: The values to bound the profile to

        :return: None

        """
        self._prof_bounds.append([var, comp, profile_range, bounds])

    def bound_point(self, var, bounds, comp=None, point=None):
        """function that allows the user to bound a certain profile to some value

        :param GeneralVar var: The pyomo variable that we will bound
        :param str comp:The component that bound applies to
        :param tuple profile_range: The range within the set to be bounded
        :param tuple bounds: The values to bound the profile to

        :return: None

        """
        if var != 'S':
            self._point_times.append(point)
        else:
            self._point_wavelengths.append(point)
        
        self._prof_points.append([var, comp, point, bounds])

    def _validate_data(self, model):
        """Verify all inputs to the model make sense.

        This method is not suppose to be used by users. Only for developers use

        :param ConcreteModel model: The created Pyomo model

        :return: None

        """
        if not self.template_component_data.component_set('concentration'):
            warnings.warn('The Model does not have any mixture components')
        else:
            if self._odes:
                if isinstance(self._odes, dict):
                    dummy_balances = self._odes
                else:
                    dummy_balances = self._odes(model, model.start_time)
                if len(self._component_names) + len(self._complementary_states) != len(dummy_balances):
                    print(
                        'WARNING: The number of ODEs is not the same as the number of state variables.\n If this is the desired behavior, some odes must be added after the model is created.')

            else:
                print(
                    'WARNING: differential expressions not specified. Must be specified by user after creating the model')

        if self._algebraics:
            if self._algebraic_constraints and not isinstance(self._algebraic_constraints, dict):
                dummy_balances = self._algebraic_constraints(model, model.start_time)
                if len(self._algebraics) != len(dummy_balances):
                    print(
                        'WARNING: The number of algebraic equations is not the same as the number of algebraic variables.\n If this is the desired behavior, some algebraics must be added after the model is created.')
            else:
                print(
                    'WARNING: algebraic expressions not specified. Must be specified by user after creating the model')

        if self._absorption_data is not None:
            if not self._meas_times:
                raise RuntimeError('Need to add measurement times')


    def _add_algebraic_var(self, model):
        """If algebraics are present, add the algebraic variable to the model

        :param ConcreteModel model: The created Pyomo model

        :return: None

        """
        if hasattr(self, 'template_algebraic_data'):    
            a_info = self.template_algebraic_data
        
            setattr(model, self.__var.algebraic, Var(model.alltime,
                                                     a_info.names,
                                                     initialize=a_info.as_dict('value')))
                                                     
            for alg in a_info:
                if alg.bounds is not None:
                    for t in model.alltime:
                        getattr(model, self.__var.algebraic)[t, alg.name].setlb(alg.lb)
                        getattr(model, self.__var.algebraic)[t, alg.name].setub(alg.ub)
                        
            # for alg in a_info:
            #     if alg.step is not None:
                     
            #     setattr(model, self.__var.step_variable, Var(model.alltime,
            #                                                  steps,
            #                                                  initialize=1.0))
                
            #     #Constraint to keep track of the step values (for diagnostics)
            #     def rule_step_function(m, t, k):
            #         step_const = getattr(m, self.__var.step_variable)[t, k] - step_fun(m, t, num=k, **self._step_data[k])
            #         return step_const == 0.0
    
            #     constraint_suffix = 'constraint'
                
            #     setattr(model, self.__var.step_variable + constraint_suffix,
            #             Constraint(model.alltime,
            #                        steps,
            #                        rule=rule_step_function))
    
        return None
    
    def _add_initial_conditions(self, model):
        """Set up the initial conditions for the model

        :param ConcreteModel model: The created Pyomo model

        :return: None
        """
        var_info = dict(self.template_component_data._dict)
        
        if hasattr(self, 'template_state_data'):
            var_info.update(self.template_state_data._dict)
        
        model.init_conditions = Var(model.states,
                                    initialize={k: v.value for k, v in var_info.items()},
                                    )
        unknown_init = {}
        
        for var, obj in model.init_conditions.items():
            if var_info[var].known:
                obj.fix()
            else:
                lb = var_info[var].lb
                ub = var_info[var].ub
                init = var_info[var].value
                obj.setlb(lb)
                obj.setub(ub)
                unknown_init[var] = init
        
        if len(unknown_init) > 0:
            model._unknown_init_set = Set(initialize=list(unknown_init.keys()))
            setattr(model, self.__var.concentration_init, Var(model._unknown_init_set,
                                                             initialize=unknown_init))
            
            model.del_component('P_all')    
            model.P_all = Set(initialize=model.parameter_names | model._unknown_init_set,
                                 ordered=True)

        def rule_init_conditions(m, k):
            if k in m.mixture_components:
                return getattr(m, self.__var.concentration_model)[m.start_time, k] - m.init_conditions[k] == 0
            else:
                return getattr(m, self.__var.state_model)[m.start_time, k] - m.init_conditions[k] == 0

        model.init_conditions_c = \
            Constraint(model.states, rule=rule_init_conditions)
            
        if hasattr(model, self.__var.concentration_init):
            
            def rule_Pinit_conditions(m, k):
                if k in m.mixture_components:
                    return getattr(m, self.__var.concentration_init)[k] - m.init_conditions[k] == 0
                else:
                    return getattr(m, self.__var.concentration_init)[k] - m.init_conditions[k] == 0
    
            model.Pinit_conditions_c = \
                Constraint(model._unknown_init_set, rule=rule_Pinit_conditions)

        return None

    def _add_model_variables(self, model):
        """Adds the model variables to the pyomo model

        :param ConcreteModel model: The created Pyomo model

        :return: None
        """
        c_info = self.template_component_data
        self._component_names = c_info.names
        
        var_set = [c_info]
        
        if hasattr(self, 'template_state_data'):
            s_info = self.template_state_data
            self._complementary_states = s_info.names
            var_set.append(s_info)
        
        for var_class in var_set:
        
            v_info = var_class    
        
            model_pred_var_name = {
                    'components' : [self.__var.concentration_model, model.mixture_components],
                    'states' : [self.__var.state_model, model.complementary_states],
                        }
            
            var, model_set = model_pred_var_name[var_class.attr_class_set_name]
            
            if hasattr(model_set, 'ordered_data') and len(model_set.ordered_data()) == 0:
                continue
            
            setattr(model, var, Var(model.alltime,
                                          model_set,
                                          initialize=1) 
                    )    
        
            for time, comp in getattr(model, var):
                if time == model.start_time.value:
                    
                    getattr(model, var)[time, comp].value = v_info[comp].value
                   
            setattr(model, f'd{var}dt', DerivativeVar(getattr(model, var),
                                                      wrt=model.alltime)
                    )

        fixed_var_name = {
                self.__var.concentration_measured : self._concentration_data,
                self.__var.state : self._complementary_states_data,
                self.__var.user_defined : self._custom_data,
                    }
        
        for var, data in fixed_var_name.items():
           
            if data is None or len(data) == 0:
                continue
            
            c_dict = dict()
            if hasattr(self, f'_is_{var}_deriv'):
                if getattr(self, f'_is_{var}_deriv') == True:
                    c_bounds = (None, None)
                else:
                    c_bounds = (0.0, None)
                
                if data is not None:
                    
                    for i, row in data.iterrows():
                         c_dict.update({(i, col): float(row[col]) for col in data.columns if not np.isnan(float(row[col]))})
                
                    setattr(model, f'{var}_indx', Set(initialize=list(c_dict.keys()), ordered=True))
                    
                    setattr(model, var, Var(getattr(model, f'{var}_indx'),
                                                  bounds=c_bounds,
                                                  initialize=c_dict,
                                                  )
                            )
                    
                    for k, v in getattr(model, var).items():
                        getattr(model, var)[k].fixed = True
                
                else:
                    setattr(model, var, Var(model.allmeas_times,
                                    model.mixture_components,
                                    bounds=c_bounds,
                                    initialize=1))
        
                    for time, comp in getattr(model, var):
                        if time == model.start_time.value:
                            getattr(model, var)[time, comp].value = v_info[comp].value

        return None
    
    def _add_model_constants(self, model):
        """Add the model constants

        :param ConcreteModel model: The created Pyomo model

        :return: None

        """
        if hasattr(self, 'template_constant_data'):
        
            con_info = self.template_constant_data
            
            setattr(model, self.__var.model_constant, Var(con_info.names,
                                                            initialize=con_info.as_dict('value'),
                                                            ))    
        
            for param, obj in getattr(model, self.__var.model_constant).items():
                obj.fix()
          
        return None

    def _add_model_parameters(self, model):
        """Add the model parameters to the pyomo model

        :param ConcreteModel model: The created Pyomo model

        :return: None

        """
        if hasattr(self, 'template_parameter_data'):
            
            p_info = self.template_parameter_data
         
            # Initial parameter values
            p_values = p_info.as_dict('value')
            for param, value in p_values.items():
                if value is None:
                    if p_info[param].bounds[0] is not None and p_info[param].bounds[1] is not None:
                        p_values[param] = sum(p_info[param].bounds)/2
                else:
                    p_values[param] = value
                        
            if self._scale_parameters:
                setattr(model, self.__var.model_parameter, Var(model.parameter_names,
                                                               bounds = (0.1, 10),
                                                               initialize=1))
    
            else:
                # print(p_values)
                setattr(model, self.__var.model_parameter,
                        Var(model.parameter_names,
                            initialize=p_values))
    
            # Set the bounds
            p_bounds = p_info.as_dict('bounds')
            for k, v in p_bounds.items():
                factor = 1
                if self._scale_parameters:
                    factor = p_values[k]
                    
                #print(p_info)
    
                if p_info[k].lb is not None:
                    lb = p_info[k].lb/factor
                    getattr(model, self.__var.model_parameter)[k].setlb(lb)
    
                if p_info[k].ub is not None:
                    ub = p_info[k].ub/factor
                    getattr(model, self.__var.model_parameter)[k].setub(ub)
                    
            #for optional smoothing parameters (CS):
            if isinstance(self._smoothparameters, dict) and self._smoothparam_data is not None:
                ps_dict = dict()
                for k in self._smoothparam_data.columns:
                    for c in model.allsmooth_times:
                        ps_dict[c, k] = float(self._smoothparam_data[k][c])
    
                ps_dict2 = dict()
                for p in self._smoothparameters.keys():
                    for t in model.alltime:
                        if t in model.allsmooth_times:
                            ps_dict2[t, p] = float(ps_dict[t, p])
                        else:
                            ps_dict2[t, p] = 0.0
    
                model.Ps = Param(model.alltime, model.smoothparameter_names, initialize=ps_dict2, mutable=True, default=20.)#here just set to some value that is noc
    
            # Fix parameters declared as fixed
            p_fixed = p_info.as_dict('fixed')
            for param, v in p_fixed.items():
                getattr(model, self.__var.model_parameter)[param].fixed = v
        
        return None

    def _add_unwanted_contribution_variables(self, model):
        """Add the bounds for the unwanted contributions, if any

        :param ConcreteModel model: The created Pyomo model

        :return: None

        """
        if self._qr_bounds is not None:
            qr_bounds = self._qr_bounds
        elif self._qr_bounds is None:
            qr_bounds = None
            
        if self._qr_init is not None:
            qr_init = self._qr_init
        elif self._qr_init is None:
            qr_init = 1.0
            
        model.qr = Var(model.alltime, bounds=qr_bounds, initialize=qr_init)
        
        if self._g_bounds is not None:
            g_bounds = self._g_bounds
        elif self._g_bounds is None:
            g_bounds = None
            
        if self._g_init is not None:
            g_init = self._g_init
        elif self._g_init is None:
            g_init = 0.1
            
        setattr(model, self.__var.unwanted_contribution, Var(model.meas_lambdas, 
                                                             bounds=g_bounds, 
                                                             initialize=g_init))
        
        return None

    def change_time(self, expr_orig, c_mod, new_time, current_model):
        """Method to remove the fixed parameters from the ConcreteModel and replace them with real parameters. This
        converts the dummy variables used to generate the expressions.

        .. note::

            At the moment, this only supports one and two dimensional variables

        :param expression expr_orig: The original user generated expression
        :param dict c_mod: The dict of dummy variables
        :param float new_time: The time to replace in the variable index
        :param ConcreteModel current_model: The model containing the variables

        :return expr_new_time: The expression with the new time

        """
        expr_new_time = expr_orig
        var_dict = c_mod
        
        for model_var, obj_list in var_dict.items():
            if not isinstance(obj_list[1].index(), int):
                old_var = obj_list[1]
                new_var = getattr(current_model, obj_list[0])[new_time, model_var]
        
            else:
                old_var = obj_list[1]
                new_var = getattr(current_model, obj_list[0])[model_var]
        
            expr_new_time = self._update_expression(expr_new_time, old_var, new_var)
    
        return expr_new_time

    @staticmethod
    def _update_expression(expr, replacement_param, change_value):
        """Takes the non-estiambale parameter and replaces it with its intitial
        value
        
        :param expression expr: the target ode constraint
        :param str replacement_param: the non-estimable parameter to replace
        :param float change_value: initial value for the above parameter
            
        :return: expression new_expr: Updated constraints with the desired parameter replaced with a float
        
        """
        visitor = ReplacementVisitor()
        visitor.change_replacement(change_value)
        visitor.change_suspect(id(replacement_param))
        new_expr = visitor.dfs_postorder_stack(expr)       
        return new_expr

    def _add_model_odes(self, model):
        """Adds the ODE system to the model, if any

        :param ConcreteModel model: The created Pyomo model

        :return: None

        """
        if isinstance(self._odes, dict):
            
            def rule_odes(m, t, k):
                exprs = self._odes                
               
                if t == m.start_time.value:
                    return Constraint.Skip
                else:
                    if k in m.mixture_components:
                        if k in exprs.keys():
                            deriv_var = f'd{self.__var.concentration_model}dt'
                            final_expr = getattr(m, deriv_var)[t, k] == exprs[k].expression
                            final_expr  = self.change_time(final_expr, self.c_mod, t, m)
                            return final_expr
                        else:
                            return Constraint.Skip
                    else:
                        if k in exprs.keys():
                            deriv_var = f'd{self.__var.state_model}dt'
                            final_expr = getattr(m, deriv_var)[t, k] == exprs[k].expression
                            final_expr  = self.change_time(final_expr, self.c_mod, t, m)
                            return final_expr
                        else:
                            return Constraint.Skip

            setattr(model, self.__var.ode_constraints, Constraint(model.alltime,
                                                                  model.states,
                                                                  rule=rule_odes))
        return None
    
    def _add_algebraic_constraints(self, model):
        """Adds the algebraic constraints the model, if any

        :param ConcreteModel model: The created Pyomo model

        :return: None

        """
        if self._algebraic_constraints:
            if hasattr(self, 'reaction_dict') or hasattr(self, '_use_alg_dict') and self._use_alg_dict:
                n_alg_eqns = list(self._algebraic_constraints.keys())
                
                def rule_algebraics(m, t, k):
                    alg_const = self._algebraic_constraints[k].expression
                    alg_var = getattr(m, self.__var.algebraic)[t, k]
                    final_expr = alg_var - alg_const == 0.0
                    final_expr  = self.change_time(final_expr, self.c_mod, t, m)
                    return final_expr
    
            model.algebraic_consts = Constraint(model.alltime,
                                                n_alg_eqns,
                                                rule=rule_algebraics)
        return None
    
    def _add_objective_custom(self, model):
        """Adds the custom objectives, if any and uses the model variable
        UD for the data. This is where the custom data is stored.

        .. note::

            This only handles one addition such objective term type

        :param ConcreteModel model: The created Pyomo model

        :return: None

        """
        def custom_obj(m, t, y):
            exprs = getattr(m, self.__var.user_defined)[t, y] - getattr(m, self.__var.algebraic)[t, y]
            return exprs**2
        
        obj = 0
        var_name = []
        for index, values in model.UD.items():
            if index[1] not in self._custom_objective:
                continue
            if index[1] not in var_name:
                var_name.append(index[1])
            obj += custom_obj(model, index[0], index[1])
            
        model.custom_obj = obj
       
        return None

    def _add_spectral_variables(self, model, estimator):
        """Add D and C variables for the spectral data

        :param ConcreteModel model: The created Pyomo model
        :param bool is_simulation: Indicates if the model is for simulation or not

        :return: None

        """
        if self._spectral_data is not None:
            s_data_dict = dict()
            index_list = list(itertools.product(self._spectral_data.index, self._spectral_data.columns))
            
            for index in index_list:
                try:
                    s_data_dict[index] = float(self._spectral_data.loc[index])
                except:
                    s_data_dict[index] = float('nan')

            setattr(model, self.__var.spectra_data, Param(model.times_spectral,
                                                          model.meas_lambdas,
                                                          domain=Reals,
                                                          initialize=s_data_dict))
            
            if estimator != 'simulator':
                setattr(model, self.__var.concentration_spectra, Var(model.times_spectral,
                                                                 model.mixture_components,
                                                                 bounds=(0, None),
                                                                 initialize=1))
            
        return None

    def _check_absorbing_species(self, model):
        """Set up the appropriate S depending on absorbing species

        :param ConcreteModel model: The created Pyomo model

        :return: None

        """
        if self._absorption_data is not None:
            s_dict = dict()
            for k in self._absorption_data.columns:
                for l in self._absorption_data.index:
                    s_dict[l, k] = float(self._absorption_data[k][l])
        
        elif self._init_absorption_data is not None:
            s_dict = dict()
            for k in self._init_absorption_data.columns:
                for l in self._init_absorption_data.index:
                    s_dict[l, k] = float(self._init_absorption_data[k][l])
                    
        else:
            s_dict = 0.0
        
        if self._is_D_deriv == True:
            s_bounds = (None, None)
        else:
            s_bounds = (0.0, None)
        
        if self.has_spectral_data():    
            
            self.set_non_absorbing_species(model, self._non_absorbing, check=False)
            
            if self._is_non_abs_set:
                
                setattr(model, self.__var.spectra_species, Var(model.meas_lambdas,
                                                               model.abs_components,
                                                               bounds=s_bounds,
                                                               initialize=0.0))
            else:
                setattr(model, self.__var.spectra_species, Var(model.meas_lambdas,
                                                               model.mixture_components,
                                                               bounds=s_bounds,
                                                               initialize=s_dict))

            if self._absorption_data is not None:
                for l in model.meas_lambdas:
                    for k in model.mixture_components:
                        getattr(model, self.__var.spectra_species)[l, k].fixed = True
                        
        return None
    
    def _check_bounds_entries(self):
        """Checks the bounded variables after the model has been built.
        
        :return: None
        
        """
        for bound_set in self._prof_bounds:

            var = bound_set[0]
            comp = bound_set[1]
            profile_range = bound_set[2]
            bounds = bound_set[3]
        
            if not isinstance(var, str):
                raise RuntimeError('var argument needs to be type string')
    
            if var not in ['C', 'U', 'S']:
                raise RuntimeError('var argument needs to be either C, U, or S')
    
            if comp is not None:
                if not isinstance(comp, str):
                    raise RuntimeError('comp argument needs to be type string')
                if comp not in self.template_component_data.names:
                    raise RuntimeError('comp needs to be one of the components')
    
            if profile_range is not None:
                if not isinstance(profile_range, tuple):
                    raise RuntimeError('profile_range needs to be a tuple')
                    if profile_range[0] > profile_range[1]:
                        raise RuntimeError('profile_range[0] must be greater than profile_range[1]')
    
            if not isinstance(bounds, tuple):
                raise RuntimeError('bounds needs to be a tuple')
                
        return None
        
    def _apply_bounds_to_variables(self, model):
        """User specified bounds to certain model variables are added here

        :param ConcreteModel model: The created Pyomo model

        :return: None

        """
        for bound_set in self._prof_bounds:
            var = bound_set[0]
            component_name = bound_set[1]
            if bound_set[2] is not None:
                bound_time_start = bound_set[2][0]
                bound_time_end = bound_set[2][1]
            upper_bound = bound_set[3][1]
            lower_bound = bound_set[3][0]
            
            for time, comp in getattr(model, var):
                if component_name == comp or component_name is None:
                    if bound_set[2] is not None:
                        if time >= bound_time_start and time < bound_time_end:
                            getattr(model, var)[time, comp].setlb(lower_bound)
                            getattr(model, var)[time, comp].setub(upper_bound)
                    else:
                        getattr(model, var)[time, comp].setlb(lower_bound)
                        getattr(model, var)[time, comp].setub(upper_bound)
                        
        for bound_set in self._prof_points:
            var = bound_set[0]
            component_name = bound_set[1]
            time = bound_set[2]
            upper_bound = bound_set[3][1]
            lower_bound = bound_set[3][0]
            getattr(model, var)[time, comp].setlb(lower_bound)
            getattr(model, var)[time, comp].setub(upper_bound)
                        
        return None

    def _add_model_smoothing_parameters(self, model):
        """Adds smoothing parameters to the model, if any.
        These are optional smoothing parameter values (mutable) read from a file

        :param ConcreteModel model: The created Pyomo model

        :return: None
        
        """
        if self._smoothparam_data is not None:
            model.smoothparameter_names = Set(initialize=self._smoothparameters.keys()) #added for mutable parameters
            model.smooth_param_datatimes = Set(initialize=sorted(self._smoothparam_data.index))
            
            help_dict=dict()
            for k in self._smoothparam_data.index:
                for j in self._smoothparam_data.columns:
                    help_dict[k, j]=float(self._smoothparam_data[j][k])
            
            model.smooth_param_data = Param(self._smoothparam_data.index, sorted(self._smoothparam_data.columns), initialize=help_dict)

        return None

    def _set_up_times(self, model, start_time, end_time, estimator):
        """This method sets up the times for the model based on all of the inputs.

        :param ConcreteModel model: The created Pyomo model
        :param float start_time: The start time (0)
        :param float end_time: The end time
        :param bool is_simulation: Indicates if the model is for simulation

        :return: None

        """
        if self._times is not None:
            if start_time is None:
                start_time = self._times[0]
            if end_time is None:
                end_time = self._times[1]
        else:
            if start_time is None and end_time is None:
                try:    
                    start_time = 0
                    end_time = self.time_span_max
                except:
                    raise ValueError('A model requires a start and end time or a dataset')
        
        start_time = self.round_time(start_time)
        end_time = self.round_time(end_time)
        model_times = {}
        model_data = {}
        
        model_times['feed'] = sorted(list(set(self._feed_times)))
        model_times['spectral'] = sorted(list(set(self._spectral_data.index))) if self._spectral_data is not None else list()
        model_times['concentration'] = sorted(list(set(self._concentration_data.index))) if self._concentration_data is not None else list()
        model_times['states'] = sorted(list(set(self._complementary_states_data.index))) if self._complementary_states_data is not None else list()
        model_times['smooth'] = sorted(list(set(self._smoothparam_data.index))) if self._smoothparam_data is not None else list()
        # model_times['point'] = set(self._point_times)
    
        model_data['lambda_D'] = sorted(list(self._spectral_data.columns)) if self._spectral_data is not None else list()
        model_data['lambda_S'] = sorted(list(self._absorption_data.index)) if self._absorption_data is not None else list()
        # model_data['lambda_point'] = set(self._point_wavelengths)

        if estimator == 'p_estimator':
            model_times['uplc'] = sorted(list(set(self._huplc_data.index))) if self._huplc_data is not None else list()
            model_times['custom'] = sorted(list(set(self._custom_data.index))) if self._custom_data is not None else list()
    
        model_times['all_times'] = sorted(list(set().union(*model_times.values())))
        model_data['all_lambdas'] = sorted(list(set().union(*model_data.values())))
        
        for key, value in model_times.items():
            if key not in ['times', 'all_times']:
                setattr(model, f'times_{key}', Set(initialize=value, ordered=True))
        
        model.feed_times = Set(initialize=model_times['feed'], ordered=True)
        model.meas_lambdas = Set(initialize=model_data['all_lambdas'], ordered=True)
        
        model.allmeas_times = Set(initialize=model_times['all_times'], ordered=True)
        model.alltime = ContinuousSet(initialize=model.allmeas_times, bounds=(start_time, end_time))
        
        model.start_time = Param(initialize=start_time, domain=Reals)
        model.end_time = Param(initialize=end_time, domain=Reals)
        
        if 'uplc' in model_times:
            model.huplcmeas_times = Set(initialize=model_times['uplc'], ordered=True)
            model.huplctime = ContinuousSet(initialize=model.huplcmeas_times, bounds=(start_time, end_time))
            
        self.start_time = start_time
        self.end_time = end_time
        self.model_times = model_times
        
        return None
    
    @staticmethod
    def _check_time_inputs(time_set, start_time, end_time):
        """Checks the first and last time of a measurement to see if it's in
        the model time bounds

        :param array-like time_set: The list of all times
        :param float start_time: The start time
        :param float end_time: The end time

        :return: None

        """
        if time_set[0] < start_time:
            raise RuntimeError(f'Measurement time {time_set[0]} not within ({start_time}, {end_time})')
        if time_set[-1] > end_time:
            raise RuntimeError(f'Measurement time {time_set[-1]} not within ({start_time}, {end_time})')
    
        return None
    
    def add_step_vars(self, step_data):
        """Adds a step variable if the data is provided

        :param dict step_data: The step data for the ReactionModel

        :return None

        """
        if not hasattr(self, '_step_data'):
            self._step_data = None
        self._step_data = step_data
        return None
    
    def _add_time_steps(self, model):
        """Builds the step functions into the model using the step_fun method

        :param ConcreteModel model: The created Pyomo model

        :return: None

        """
        from kipet.model_tools.model_functions import step_fun

        # Add dosing var
        setattr(model, self.__var.dosing_variable, Var(model.alltime,
                                                      [self.__var.dosing_component],
                                                      initialize=0))
            
        def build_step_funs(m, t, k, data):
            
            step = 0
            off_modifier = 0
            for i in range(len(data)):
                if i > 0 and data[i]['switch'] == 'off':
                    off_modifier += 1
                step += step_fun(m, t, num=f'{k}_{i}', **data[i])
                
            return step - off_modifier
        
        def rule_step_function(m, t, k):
            step_const = getattr(m, self.__var.step_variable)[t, k] - build_step_funs(m, t, k, self._step_data[k])    # step_fun(m, t, num=k, **self._step_data[k])
            return step_const == 0.0
        
        if hasattr(self, '_step_data') and len(self._step_data) > 0:
            steps = list(self._step_data.keys())
            
            #for step in steps:
                #step_info = self._step_data[step]
                #num_of_steps = len(step_info)
                #step_index = list(range(num_of_steps))
                  
            tsc_index = []
            tsc_init = {}
            tsc_bounds = {}
            tsc_fixed = {}
        
            for k, v in self._step_data.items():
                for i, step_dict in enumerate(v):
                    tsc_init[f'{k}_{i}'] = step_dict['time']
                    tsc_bounds[f'{k}_{i}'] = step_dict.get('bounds', (model.start_time, model.end_time))
                    tsc_fixed[f'{k}_{i}'] = step_dict['fixed']
                    tsc_index.append(f'{k}_{i}')
                    
            
            setattr(model, self.__var.time_step_change, Var(tsc_index,
                                                           # step_index,
                    initialize=tsc_init,
                    ))
                   
            # Set the bounds
            for k, v in getattr(model, self.__var.time_step_change).items():
                v.setlb(tsc_bounds[k][0])
                v.setub(tsc_bounds[k][1])
                
                if tsc_fixed[k]:
                    v.fix()
            
            setattr(model, self.__var.step_variable, Var(model.alltime,
                                                         steps,
                                         #                step_index,
                                                         initialize=1.0))
                
            #Constraint to keep track of the step values (for diagnostics)

            constraint_suffix = 'constraint'
            
            setattr(model, self.__var.step_variable + constraint_suffix,
                    Constraint(model.alltime,
                               steps,
                         #      step_index,
                               rule=rule_step_function))   
            
            # If needed, this section may be implemented
            # def rule_step_alg(m, t, k, k2, factor):
            #       print(t, k, k2, factor)
            #       step_const = 1/factor*getattr(m, self.__var.algebraic)[t, k] - step_fun(m, t, num=k2, **self._step_data[k2])
            #       return step_const == 0.0

            # if hasattr(self, 'template_algebraic_data'):
            #     alg_steps = self.template_algebraic_data.steps
                
            #     alg_values = {alg.name: alg.value for alg in alg_steps.values()}
            #     alg_set = [alg.name for alg in alg_steps.values()]
            #     alg_step = {alg.name: alg.step for alg in alg_steps.values()}
                           
            #     steps = list(set(steps).difference(alg_set))                  
        
            #     setattr(model, self.__var.algebraic + constraint_suffix,
            #             Constraint(model.alltime,
            #                        alg_set,
            #                        alg_step.values(),
            #                        alg_values.values(),
            #                        rule=rule_step_alg,
            #                        ))
                
        return None
            
    def create_pyomo_model(self, start_time=None, end_time=None, estimator='simulator'):
        """Create a pyomo model from the provided data.

        This method is the core method for further simulation or optimization studies

        :param flaot start_time: Initial time considered in the model
        :param float end_time: Final time considered in the model
        :param bool is_simulation: Indicates if the model is for simulation

        :return: The finished Pyomo ConcreteModel
        :rtype: ConcreteModel

        """
        print(f'# TemplateBuilder: Preparing model for {estimator}')
        is_simulation = estimator == 'simulator' 
        
        if self._spectral_data is not None and self._absorption_data is not None:
            raise RuntimeError('Either add absorption data or spectral data but not both')
        
        pyomo_model = ConcreteModel()

        # Declare Sets
        pyomo_model.mixture_components = Set(initialize=self.template_component_data.names)
        
        if not hasattr(self, 'template_parameter_data'):
            parameter_names = []
        else:
            parameter_names = self.template_parameter_data.names
        pyomo_model.parameter_names = Set(initialize=parameter_names)
        
        if not hasattr(self, 'template_state_data'):
            state_names = []
        else:
            state_names = self.template_state_data.names
        pyomo_model.complementary_states = Set(initialize=state_names)
        
        pyomo_model.states = pyomo_model.mixture_components | pyomo_model.complementary_states
        pyomo_model.measured_data = Set(initialize=self._all_state_data)
        
        # Set up the model time sets and parameters
        self._set_up_times(pyomo_model, start_time, end_time, estimator)
        
        # Set up the model elements by calling the following methods:
        self._add_model_variables(pyomo_model)
        self._add_model_parameters(pyomo_model)
            
        if not hasattr(self, 'template_algebraic_data'):
            alg_names = []
        else:
            alg_names = self.template_algebraic_data.names
        pyomo_model.algebraics = Set(initialize=alg_names)
            
        self._add_algebraic_var(pyomo_model)
        self._add_model_constants(pyomo_model)
        self._add_initial_conditions(pyomo_model)
        self._add_spectral_variables(pyomo_model, estimator)
        
        # This may be added below (no examples of this)
        self._add_model_smoothing_parameters(pyomo_model)
        
        # If unwanted contributions are being handled:
        if self._G_contribution is not None:
            self._add_unwanted_contribution_variables(pyomo_model)

         # Add time step variables
        self._add_time_steps(pyomo_model)

        if not hasattr(self, 'c_mod'):
            self.c_mod = Comp(pyomo_model)
            
        # Exit point for the dummy model
        if hasattr(self, 'early_return') and self.early_return:
            return pyomo_model

        # Validate the model before writing constraints
        self._validate_data(pyomo_model)

        # Add constraints
        self._add_model_odes(pyomo_model)
        self._add_algebraic_constraints(pyomo_model)
        
        # Add objective terms for custom data
        if self._custom_objective and estimator == 'p_estimator':
            self._add_objective_custom(pyomo_model)
        
        # Check the absorbing species sets
        if not is_simulation:
            self._check_absorbing_species(pyomo_model)
        
        # Add bounds, is specified
        if not is_simulation:
            self._apply_bounds_to_variables(pyomo_model)
       
        # Add given state standard deviations to the pyomo model
        self._state_sigmas = self.template_component_data.var_variances()
        if hasattr(self, 'template_state_data'):
            self._state_sigmas.update(**self.template_state_data.var_variances())
        
        # Add state/component variances
        for k, v in self._state_sigmas.items():
            if v is None:
                self._state_sigmas[k] = 1
       
        state_sigmas = {k: v for k, v in self._state_sigmas.items() if k in pyomo_model.measured_data}
        pyomo_model.sigma = Param(
            pyomo_model.measured_data, 
            domain=Reals, 
            initialize=state_sigmas
        )
       
        # Add UPLC data too the ParameterEstimator model
        if estimator ==  'p_estimator':
            if self._huplc_data is not None and self._is_huplc_abs_set:
                self.set_huplc_absorbing_species(pyomo_model, self._huplc_absorbing, self._vol, self._solid_spec, check=False)

        # In case of a second call after known_absorbing has been declared
        
        # This needs to be handled like a normal update
        # If a species has .S that is not None, then this should be called here
        
        if not is_simulation:
        
            known_abs = []
            known_abs_data = {}
            for comp in self.template_component_data:
                if comp.S is not None:
                    known_abs.append(comp.name)
                    known_abs_data[comp.name] = comp.S
            
            self._known_absorbance = known_abs
            self._known_absorbance_data = known_abs_data
            
            if len(self._known_absorbance) > 0:  
                self.set_known_absorbing_species(
                    pyomo_model,
                    self._known_absorbance,
                    self._known_absorbance_data,
                    check=False
                )
            
        return pyomo_model

    @property
    def num_parameters(self):
        """Returns the number of parameters

        :return: The number of paramters
        :rtype: float

        """
        return len(self._parameters)


    @property
    def num_mixture_components(self):
        """Returns the number of components

        :return: The number of components
        :rtype: float

        """
        return len(self._component_names)


    @property
    def num_complementary_states(self):
        """Returns the number of states

        :return: The number of states
        :rtype: float

        """
        return len(self._complementary_states)


    @property
    def num_algebraics(self):
        """Returns the number of algebraics

        :return: The number of algebraics
        :rtype: float

        """
        return len(self._algebraics)


    @property
    def measurement_times(self):
        """Returns the measurement times

        :return: The measurement times
        :rtype: float

        """
        return self._meas_times


    @property
    def huplcmeasurement_times(self):
        """Returns the HUPLC measurement times

        :return: The number of HUPLC measurement times
        :rtype: float

        """
        return self._huplcmeas_times


    @property
    def feed_times(self):
        """Returns the feed times

        :return: The feed times
        :rtype: float

        """
        return self._feed_times  # added for feeding points CS

    def has_spectral_data(self):
        """Return bool if spectral data is in the model

        :return: True if spectral data is in the model

        """
        return self._spectral_data is not None

    def has_adsorption_data(self):
        """Return bool if adsorption data is in the model

        :return: True if adsorption data is in the model

        """
        return self._absorption_data is not None

    def has_concentration_data(self):
        """Return bool if concentration data is in the model

        :return: True if concentration data is in the model

        """
        return self._concentration_data is not None

    def optional_warmstart(self, model):
        """Updates suffixes with previous results

        :param ConcreteModel model: The current Pyomo model

        :return: None

        """
        if hasattr(model, 'dual') and hasattr(model, 'ipopt_zL_out') and hasattr(model, 'ipopt_zU_out') and hasattr(model, 'ipopt_zL_in') and hasattr(model, 'ipopt_zU_in'):
            print('warmstart model components already set up before.')
        else:
            model.dual = Suffix(direction=Suffix.IMPORT_EXPORT)
            model.ipopt_zL_out = Suffix(direction=Suffix.IMPORT)
            model.ipopt_zU_out = Suffix(direction=Suffix.IMPORT)
            model.ipopt_zL_in = Suffix(direction=Suffix.EXPORT)
            model.ipopt_zU_in = Suffix(direction=Suffix.EXPORT)

    def set_non_absorbing_species(self, model, non_abs_list, check=True):
        """Sets the non absorbing component of the model.
        
        .. note::
            
            This could be simplified by making Cs always exist. This would
            remove quite a few if statements throughout the code.

        :param list non_abs_list: List of non absorbing components.
        :param ConcreteModel model: The corresponding model.
        :param bool check: Safeguard against setting this up twice.

        :return: None

        """
        # if hasattr(model, 'non_absorbing'):
        #     print("non-absorbing species were already set up before.")
        #     return

        # if (self._is_non_abs_set and check):
        #     raise RuntimeError('Non absorbing species have been already set up.')
        if hasattr(model, 'abs_components'):
            model.del_component('abs_components')
        
        self._is_non_abs_set = True
        self._non_absorbing = [] if non_abs_list is None else non_abs_list
        #model.add_component('non_absorbing', Set(initialize=self._non_absorbing))

        # Exclude non absorbing species from S matrix and create subset Cs of C:
        #model.add_component('abs_components_names', Set())
        
        set_all_components = set(model.mixture_components)
        set_non_abs_components = set(self._non_absorbing)
        set_abs_components = set_all_components.difference(set_non_abs_components)
        
        list_abs_components = sorted(list(set_abs_components))
        
        model.add_component('abs_components', Set(initialize=list_abs_components))
        #model.add_component('Cs', Var(model.times_spectral, list_abs_components))
        #model.add_component('abs_subset_contraint', ConstraintList())
        
        # for time in model.times_spectral:
        #     for comp in model.abs_components:
        #         model.abs_subset_contraint.add(model.Cs[time, comp] == model.C[time, comp])
                    
        return None

    def set_known_absorbing_species(self, model, known_abs_list, absorbance_data, check=True):
        """Sets the known absorbance profiles for specific components of the model.

        :param list known_abs_list: List of known species absorbance components.
        :param ConcreteModel model: The corresponding model.
        :param pandas.DataFrame absorbance_data: the dataframe containing the known component spectra
        :param bool check: Safeguard against setting this up twice.

        :return: None

        """
        if hasattr(model, 'known_absorbance'):
            print("species with known absorbance were already set up before.")
            return

        if (self._is_non_abs_set and check):
            raise RuntimeError('Species with known absorbance were already set up before.')

        self._is_known_abs_set = True
        model.add_component('known_absorbance', Set(initialize=self._known_absorbance))
        S = getattr(model, self.__var.spectra_species)
        lambdas = getattr(model, 'meas_lambdas')
        model.known_absorbance_data = self._known_absorbance_data
        
        for component in self._known_absorbance:
            for l in lambdas:
                S[l, component].set_value(self._known_absorbance_data[component].loc[l, component])
                S[l, component].fix()
                
        return None

    def set_huplc_absorbing_species(self, model, huplc_abs_list, vol=None, solid_spec=None, check=True):
        """Sets the non absorbing component of the model.

        :param list huplc_abs_list: List of huplc absorbing components.
        :param float vol: reactor volume
        :param str solid_spec: solid species observed with huplc data
        :param ConcreteModel model: The corresponding model.
        :param bool check: Safeguard against setting this up twice.

        :return: None

        """
        if hasattr(model, 'huplc_absorbing'):
            print("huplc-absorbing species were already set up before.")
            return

        if (self._is_huplc_abs_set and check):
            raise RuntimeError('huplc-absorbing species have been already set up.')

        self._is_huplc_abs_set = True
        #self._vol=vol
        #model.add_component('vol', Param(initialize=self._vol, within=Reals))

        self._huplc_absorbing = huplc_abs_list
        model.add_component('huplc_absorbing', Set(initialize=self._huplc_absorbing))
        times = getattr(model, 'huplcmeas_times')
        timesh = getattr(model, 'huplctime')

        model.add_component('huplcabs_components_names', Set())
        huplcabscompsnames = [name for name in set(sorted(self._huplc_absorbing))]
        model.add_component('huplcabs_components', Set(initialize=huplcabscompsnames))

        self._solid_spec=solid_spec
        if solid_spec is not None:
            solid_spec_arg1keys = [key for key in set(sorted(self._solid_spec.keys()))]
            model.add_component('solid_spec_arg1', Set(initialize=solid_spec_arg1keys))
            solid_spec_arg2keys = [key for key in set(sorted(self._solid_spec.values()))]
            model.add_component('solid_spec_arg2', Set(initialize=solid_spec_arg2keys))
            model.add_component('solid_spec', Var(model.solid_spec_arg1, model.solid_spec_arg2, initialize=self._solid_spec))

        Dhat_dict=dict()
        for k in self._huplc_data.columns:
            for c in self._huplc_data.index:
                Dhat_dict[c, k] = float(self._huplc_data[k][c])

        model.add_component('fixed_Dhat', ConstraintList())
        model.add_component('Dhat', Var(times, huplcabscompsnames,initialize=Dhat_dict))
        model.add_component('Chat', Var(timesh, huplcabscompsnames,initialize=1))
        model.add_component('matchChatZ', ConstraintList())
        
        return None



"""
Effort to remove the PyomoSimulator class
"""
# Standard library imports

# Third library import
import numpy as np
import pandas as pd
from pyomo.environ import Suffix
from pyomo.core.base import TransformationFactory
from pyomo.opt import SolverFactory

# KIPET library imports
from kipet.calculation_tools.interpolation import interpolate_trajectory
from kipet.model_tools.visitor_classes import ScalingVisitor
from kipet.estimator_tools.results_object import ResultsObject
from kipet.model_tools.pyomo_model_tools import (get_index_sets,
                                                 index_set_info,
                                                 model_info)
from kipet.general_settings.variable_names import VariableNames

__var = VariableNames()

"""Simulator based on pyomo.dae discretization strategies.

:Methods:
    
    - :func:`apply_discretization`
    - :func:`scale_model`
    - :func:`scale_parameters`
    - :func:`scale_expression`
    - :func:`fix_from_trajectory`
    - :func:`unfix_time_dependent_variable`
    - :func:`initialize_parameters`
    - :func:`build_set_new`
    - :func:`initialize_from_trajectory`
    - :fuch:`scale_variables_from_trajectory`
    - :func:`run_sim`
    - :func:`add_warm_start_suffixes`
    - :func:`update_warm_start`

"""


# def __init__(self, model):
#     """Simulator constructor.

#     :param ConcreteModel model: The Pyomo model of the ReactionModel

#     """
#     self.__var = VariableNames()
#     self.model = model

#     # Most of the attributes are model attributes
#     self.attrs = model_info(self.model)
#     for key, value in self.attrs.items():
#         setattr(self, f'_{key}', value)

#     # Creates a scaling factor suffix
#     if not hasattr(self.model, 'scaling_factor'):
#         self.model.scaling_factor = Suffix(direction=Suffix.EXPORT)

def apply_discretization(model, transformation, **kwargs):
    """Discretizes the model.

    :param str transformation: The type of transformation (only dae.collocation...)
    :param dict kwargs: The options for the discretization

    :return: None

    """
    fixed_times = kwargs.pop('fixed_times', None)

    if not model.alltime.get_discretization_info():

        discretizer = TransformationFactory(transformation)

        if fixed_times == None:
            discretizer.apply_to(model, wrt=model.alltime, **kwargs)
        else:
            discretizer.apply_to(model, wrt=fixed_times, **kwargs)

        alltimes = sorted(model.alltime)
        n_alltimes = len(alltimes)

        # This needs to be looked at in more detail to see if it is still needed.

        # added for optional smoothing parameter with reading values from file CS:
        # if self._smoothparam_given:
        #     dfps = pd.DataFrame(index=self.model.alltime, columns=self.model.smoothparameter_names)
        #     for t in self.model.alltime:
        #         if t not in self.model.allsmooth_times:  # for points that are the same in original meas times and feed times
        #             dfps.loc[t] = float(22.) #something that is not between 0 and 1
        #         else:
        #             ps_dict_help = dict()
        #             for p in self.model.smoothparameter_names:
        #                 ps_dict_help[t, p] = value(self.model.smooth_param_data[t, p])
        #             dfps.loc[t] = [ps_dict_help[t, p] for p in self.model.smoothparameter_names]
        #     dfallps = dfps
        #     dfallps.sort_index(inplace=True)
        #     dfallps.index = dfallps.index.to_series().apply(
        #         lambda x: np.round(x, 6))  # time from data rounded to 6 digits

        #     dfallpsall = pd.DataFrame(index=self.model.alltime, columns=self.model.smoothparameter_names)
        #     dfsmoothdata = pd.DataFrame(index=sorted(self.model.smooth_param_datatimes), columns=self.model.smoothparameter_names)

        #     for t in self.model.smooth_param_datatimes:
        #         dfsmoothdata.loc[t] = [value(self.model.smooth_param_data[t, p]) for p in self.model.smoothparameter_names]

        #     for p in self.model.smoothparameter_names:
        #         values = interpolate_trajectory(self.model.alltime, dfsmoothdata[p])
        #         for i, ti in enumerate(self.model.alltime):
        #             if float(dfallps[p][ti]) > 1:
        #                 valueinterp=values[i]
        #                 dfallpsall[p][ti] = float(valueinterp)
        #             else:
        #                 dfallpsall.loc[ti] = float(dfallps[p][ti])

        _default_initialization(model)

        # if hasattr(self.model, 'K'):
        #     print('Scaling the parameters')
        #     self.scale_parameters()

    else:
        print('***WARNING: Model already discretized. Ignoring second discretization')


def scale_model(model):
    """Method to scale the model parameters

    :return: None

    """
    if hasattr(model, __var.model_parameter_scaled):
        print('Scaling the parameters')
        scale_parameters()

    return None


def scale_parameters(model):
    """If scaling, this multiplies the constants in model.K to each
    parameter in model.P.
    
    I am not sure if this is necessary and will look into its importance.

    :return: None

    """
    scale = {}
    for i in model.P:
        scale[id(model.P[i])] = model.K[i]

    for var in __var.modeled_states:
        for i in getattr(model, var):
            scale[id(getattr(model, var)[i])] == 1

    for k, v in getattr(model, __var.ode_constraints).items():
        scaled_expr = scale_expression(v.body, scale)
        # self.model.odes[k] = scaled_expr == 0
        getattr(model, __var.ode_constraints)[k] = scaled_expr == 0


def scale_expression(expr, scale):
    """Replace variables in an expression with scaled variables

    :param expression expr: The target expression
    :param dict scale: The mapping of scale factors to the variables

    :return: The expression after the scaled variables have been updated

    """
    visitor = ScalingVisitor(scale)
    return visitor.dfs_postorder_stack(expr)


def fix_from_trajectory(model, variable_name, variable_index, trajectories):
    """Takes in data and fixes a trajectory to the data (interpolates if necessary)

    :param str variable_name: The name of the variable type (Y, Z, etc.)
    :param str variable_index: The name of the variable (A, B, etc.)
    :param pandas.DataFrame trajectories: The dataset containing the target trajectory

    :return: The list of fixed values
    :rtype: list

    """
    if variable_name in __var.modeled_states:
        pass
        # raise NotImplementedError("Fixing state variables is not allowd. Only algebraics can be fixed")

    single_traj = trajectories[variable_index]
    sim_alltimes = sorted(model.alltimes)
    var = getattr(model, variable_name)
    values = interpolate_trajectory(sim_alltimes, single_traj)
    r_values = []
    for i, t in enumerate(sim_alltimes):
        var[t, variable_index].fix(values[i])
        r_values.append((t, values[i]))

    return r_values


def unfix_time_dependent_variable(model, variable_name, variable_index):
    """Sets the fixed attribute of a model variable to False
    
    :param ConcreteModel model: A Pyomo model
    :param str variable_name: The name of the variable to unfix
    :param str variable_index: The specific component of the variable
    
    :return: None
    
    """
    var = getattr(model, variable_name)
    sim_times = sorted(model.alltimes)
    for i, t in enumerate(sim_times):
        var[t, variable_index].fixed = False


def _default_initialization(model):
    """Initializes discreted variables model with initial condition values.

    This method is not intended to be used by users directly

    :return: None

    """
    tol = 1e-4
    attrs = model_info(model)

    if hasattr(model, __var.concentration_model):
        z_init = []
        for t in attrs['alltimes']:
            for k in attrs['mixture_components']:
                if abs(model.init_conditions[k].value) > tol:
                    z_init.append(model.init_conditions[k].value)
                else:
                    z_init.append(1.0)

        z_array = np.array(z_init).reshape((attrs['n_alltimes'], attrs['n_components']))
        z_init_panel = pd.DataFrame(data=z_array,
                                    columns=attrs['mixture_components'],
                                    index=attrs['alltimes'])

        initialize_from_trajectory(model, __var.concentration_model, z_init_panel)

    c_init = []
    if attrs['concentration_given']:
        pass
    else:
        for t in attrs['allmeas_times']:
            for k in attrs['mixture_components']:
                if t in attrs['meas_times']:
                    if abs(model.init_conditions[k].value) > tol:
                        c_init.append(model.init_conditions[k].value)
                    else:
                        c_init.append(1.0)
                else:
                    c_init.append(float('nan'))  # added for new huplc structure!

    if attrs['n_allmeas_times']:
        if attrs['concentration_given']:
            pass
        else:
            c_array = np.array(c_init).reshape((attrs['n_allmeas_times'], attrs['n_components']))
            c_init_panel = pd.DataFrame(data=c_array,
                                        columns=attrs['mixture_components'],
                                        index=attrs['allmeas_times'])

            if hasattr(model, __var.concentration_measured):
                initialize_from_trajectory(model, __var.concentration_measured, c_init_panel)
                print("self._n_meas_times is true in _default_init in PyomoSim")

    if hasattr(model, __var.state_model):
        x_init = []
        for t in attrs['alltimes']:
            for k in attrs['complementary_states']:
                if abs(model.init_conditions[k].value) > tol:
                    x_init.append(model.init_conditions[k].value)
                else:
                    x_init.append(1.0)

        x_array = np.array(x_init).reshape((attrs['n_alltimes'], attrs['n_complementary_states']))
        x_init_panel = pd.DataFrame(data=x_array,
                                    columns=attrs['complementary_states'],
                                    index=attrs['alltimes'])

        initialize_from_trajectory(model, __var.state_model, x_init_panel)


def initialize_parameters(model, params):
    """Initialize the parameters given a dict of parameter values

    :param dict params: A dictionary with parameter keys and initial values

    :return: None

    """
    for k, v in params.items():
        getattr(model, __var.model_parameter).value = v


def build_sets_new(model, variable_name, trajectories):
    """Checks trajectories to see if they have the correct format

    :param str variable_name: The model variable
    :param pandas.DataFrame trajectories: The trajectory data

    :return: the inner and component sets as a tuple
    :rtype: tuple

    """
    var = getattr(model, variable_name)
    index_sets = get_index_sets(var)

    if isinstance(trajectories, pd.DataFrame):
        if variable_name not in __var.__dict__.values():
            inner_set = list(trajectories.index)
            component_set = list(trajectories.columns)
        else:

            if var.dim() > 1 and var.index_set().dim() == 0:
                inner_set = list(set([t[0] for t in var.index_set()]))
                component_set = list(set([t[1] for t in var.index_set()]))
                inner_set.sort()
            else:
                index_sets = get_index_sets(var)
                index_dict = index_set_info(index_sets)

                inner_set = getattr(model, index_sets[index_dict['cont_set'][0]].name)
                component_set = getattr(model, index_sets[index_dict['other_set'][0]].name)

        return inner_set, component_set

    else:
        return None, None


def initialize_from_trajectory(model, variable_name, trajectories):
    """Initializes discretized points with values from trajectories.

    :param str variable_name : Name of the variable in pyomo model
    :param pandas.DataFrame trajectories: Indexed in in the same way the pyomo
        variable is indexed. If the variable is by two sets then the first set is
        the indices of the data frame, the second set is the columns

    :return: None

    """
    if not model.alltime.get_discretization_info():
        raise RuntimeError('apply discretization first before initializing')

    var = getattr(model, variable_name)
    inner_set, component_set = build_sets_new(model, variable_name, trajectories)

    if inner_set is None and component_set is None:
        return None

    for component in component_set:
        if component in trajectories.columns:
            single_trajectory = trajectories[component]
            values = interpolate_trajectory(inner_set, single_trajectory)
            for i, t in enumerate(inner_set):
                if not np.isnan(values[i]):
                    var[t, component].value = values[i]

    return None


def scale_variables_from_trajectory(model, variable_name, trajectories):
    """Scales discretized variables with maximum value of the trajectory.

    .. note::

        This method only works with ipopt

    :param str variable_name : Name of the variable in pyomo model
    :param pandas.DataFrame trajectories: Indexed in in the same way the pyomo
        variable is indexed. If the variable is by two sets then the first set is
        the indices of the data frame, the second set is the columns

    :return: None

    """
    tol = 1e-5

    var = getattr(model, variable_name)
    inner_set, component_set = build_sets_new(model, variable_name, trajectories)

    if inner_set is None and component_set is None:
        return None

    for component in component_set:

        if component not in component_set:
            raise RuntimeError(f'Component {component} is not used in the model')

        nominal_vals = abs(trajectories[component].max())
        if nominal_vals >= tol:
            scale = 1.0 / nominal_vals
            for t in inner_set:
                model.scaling_factor.set_value(var[t, component], scale)

    return None


def run_sim(model, solver, **kwds):
    """ Runs simulation by solving nonlinear system with IPOPT

    :param str solver: The name of the nonlinear solver to used
    :param dict kwds: A dict of options passed to the solver

    :Keyword Args:

        - solver_opts (dict, optional): Options passed to the nonlinear solver
        - variances (dict, optional): Map of component name to noise variance. The
          map also contains the device noise variance
        - tee (bool,optional): flag to tell the simulator whether to stream output
          to the terminal or not

    :return: None

    """
    solver_opts = kwds.pop('solver_opts', dict())
    sigmas = kwds.pop('variances', dict())
    tee = kwds.pop('tee', False)
    seed = kwds.pop('seed', None)

    if not model.alltime.get_discretization_info():
        raise RuntimeError('apply discretization first before runing simulation')

    # Adjusts the seed to reproduce results with noise
    np.random.seed(seed)

    # Variables
    # Z_var = self.model.Z
    # if hasattr(self.model, 'Cm'):
    #     C_var = self.model.Cm  # added for estimation with inputs and conc data CS

    # if self._huplc_given: #added for additional data CS
    #     Dhat_var = self.model.Dhat
    #     Chat_var = self.model.Chat

    # # Deactivates objective functions for simulation
    # if self.model.nobjectives():
    #     objectives_map = self.model.component_map(ctype=Objective, active=True)
    #     active_objectives_names = []
    #     for obj in objectives_map.values():
    #         name = obj.getname()
    #         active_objectives_names.append(name)
    #         str_warning = 'Deactivating objective {} for simulation'.format(name)
    #         warnings.warn(str_warning)
    #         obj.deactivate()

    opt = SolverFactory(solver)

    for key, val in solver_opts.items():
        opt.options[key] = val

    solver_results = opt.solve(model, tee=tee, symbolic_solver_labels=True)
    results = ResultsObject()

    # activates objective functions that were deactivated
    # if self.model.nobjectives():
    #     active_objectives_names = []
    #     objectives_map = self.model.component_map(ctype=Objective)
    #     for name in active_objectives_names:
    #         objectives_map[name].activate()

    # retriving solutions to results object
    results.load_from_pyomo_model(model)

    return results
    # c_noise_results = []

    # w = np.zeros((self._n_components, self._n_allmeas_times))
    # n_sig = np.zeros((self._n_components, self._n_allmeas_times))
    # # for the noise term
    # if sigmas:
    #     for i, k in enumerate(self._mixture_components):
    #         if k in sigmas.keys():
    #             sigma = sigmas[k] ** 0.5
    #             dw_k = np.random.normal(0.0, sigma, self._n_allmeas_times)
    #             n_sig[i, :] = np.random.normal(0.0, sigma, self._n_allmeas_times)
    #             w[i, :] = np.cumsum(dw_k)

    # # this addition is not efficient but it can be changed later

    # for i, t in enumerate(self._allmeas_times):
    #     for j, k in enumerate(self._mixture_components):
    #         # c_noise_results.append(Z_var[t,k].value+ w[j,i])
    #         c_noise_results.append(Z_var[t, k].value + n_sig[j, i])

    # c_noise_array = np.array(c_noise_results).reshape((self._n_allmeas_times, self._n_components))
    # results.C = pd.DataFrame(data=c_noise_array,
    #                           columns=self._mixture_components,
    #                           index=self._allmeas_times)

    # #added due to new structure for non_abs species, Cs as subset of C (CS):
    # if hasattr(self, '_abs_components'):
    #     cs_noise_results=[]
    #     for i, t in enumerate(self._allmeas_times):
    #         # if i in self._meas_times:
    #         for j, k in enumerate(self._abs_components):
    #             # c_noise_results.append(Z_var[t,k].value+ w[j,i])
    #             cs_noise_results.append(Z_var[t, k].value + n_sig[j, i])

    #     cs_noise_array = np.array(cs_noise_results).reshape((self._n_allmeas_times, self._nabs_components))
    #     results.Cs = pd.DataFrame(data=cs_noise_array,
    #                               columns=self._abs_components,
    #                               index=self._allmeas_times)


#        addition for inputs estimation with concentration data CS:
# if self._concentration_given == True and self._absorption_given == False:
#     c_noise_results = []
#     for i, t in enumerate(self._allmeas_times):
#         # if i in self._meas_times:
#         for j, k in enumerate(self._mixture_components):
#             c_noise_results.append(C_var[t, k].value)
#     c_noise_array = np.array(c_noise_results).reshape((self._n_allmeas_times, self._n_components))
#     results.C = pd.DataFrame(data=c_noise_array,
#                               columns=self._mixture_components,
#                               index=self._allmeas_times)

# if self._huplc_given == True:
#     results.load_from_pyomo_model(self.model,
#                               to_load=['Chat'])
# s_results = []
# # added due to new structure for non_abs species, non-absorbing species not included in S (CS):
# if hasattr(self, '_abs_components'):
#     for l in self._meas_lambdas:
#         for k in self._abs_components:
#             s_results.append(self.model.S[l, k].value)
# else:
#     for l in self._meas_lambdas:
#         for k in self._mixture_components:
#             s_results.append(self.model.S[l, k].value)

# d_results = []
# if sigmas:
#     sigma_d = sigmas.get('device') ** 0.5 if "device" in sigmas.keys() else 0
# else:
#     sigma_d = 0
# if s_results and c_noise_results:
#     # added due to new structure for non_abs species, Cs and S as above(CS):
#     if hasattr(self,'_abs_components'):
#         for i, t in enumerate(self._meas_times):
#             #if t in self._meas_times:
#                 # print(i, t)
#             for j, l in enumerate(self._meas_lambdas):
#                 suma = 0.0
#                 for w, k in enumerate(self._abs_components):
#                     # print(i, self._meas_times)
#                     Cs = cs_noise_results[i * self._nabs_components + w]
#                     S = s_results[j * self._nabs_components + w]
#                     suma += Cs * S
#                 if sigma_d:
#                     suma += np.random.normal(0.0, sigma_d)
#                 d_results.append(suma)
#                 # print(d_results)
#     else:
#         for i, t in enumerate(self._meas_times):
#             # # print(i, t)
#             # if t in self._meas_times:
#             for j, l in enumerate(self._meas_lambdas):
#                 suma = 0.0
#                 for w, k in enumerate(self._mixture_components):
#                     # print(i, self._meas_times)
#                     C = c_noise_results[i * self._n_components + w]
#                     S = s_results[j * self._n_components + w]
#                     suma += C * S
#                 if sigma_d:
#                     suma += np.random.normal(0.0, sigma_d)
#                 d_results.append(suma)
#                 # print(d_results)
# # added due to new structure for non_abs species, non-absorbing species not included in S (CS):
# if hasattr(self, '_abs_components'):
#     s_array = np.array(s_results).reshape((self._n_meas_lambdas, self._nabs_components))
#     results.S = pd.DataFrame(data=s_array,
#                               columns=self._abs_components,
#                               index=self._meas_lambdas)
# else:
#     s_array = np.array(s_results).reshape((self._n_meas_lambdas, self._n_components))
#     results.S = pd.DataFrame(data=s_array,
#                               columns=self._mixture_components,
#                               index=self._meas_lambdas)


# d_array = np.array(d_results).reshape((self._n_meas_times, self._n_meas_lambdas))
# results.D = pd.DataFrame(data=d_array,
#                          columns=self._meas_lambdas,
#                          index=self._meas_times)

# s_data_dict = dict()
# for t in self._meas_times:
#     for l in self._meas_lambdas:
#         s_data_dict[t, l] = float(results.D[l][t])

# # Added due to estimation with fe-factory and inputs where data already loaded to model before (CS)
# if self._spectra_given:
#     self.model.del_component(self.model.D)
#     self.model.del_component(self.model.D_index)
# #########

# self.model.D = Param(self._meas_times,
#                      self._meas_lambdas,
#                      initialize=s_data_dict)

# param_vals = dict()
# for name in self.model.parameter_names:
#     param_vals[name] = self.model.P[name].value

# results.P = param_vals

def add_warm_start_suffixes(model, use_k_aug=False):
    """Adds suffixed variables to problem

    :param ConcreteModel model: A Pyomo model
    :param bool use_k_aug: Indicates if k_aug solver is being used

    :return: None

    """
    # Ipopt bound multipliers (obtained from solution)
    model.ipopt_zL_out = Suffix(direction=Suffix.IMPORT)
    model.ipopt_zU_out = Suffix(direction=Suffix.IMPORT)
    # Ipopt bound multipliers (sent to solver)
    model.ipopt_zL_in = Suffix(direction=Suffix.EXPORT)
    model.ipopt_zU_in = Suffix(direction=Suffix.EXPORT)
    # Obtain dual solutions from first solve and send to warm start
    model.dual = Suffix(direction=Suffix.IMPORT_EXPORT)

    if use_k_aug:
        model.dof_v = Suffix(direction=Suffix.EXPORT)
        model.rh_name = Suffix(direction=Suffix.IMPORT)

    return None


def update_warm_start(model):
    """Updates the suffixed variables for a warmstart

    :param ConcreteModel model: A Pyomo model

    :return: None

    """
    model.ipopt_zL_in.update(model.ipopt_zL_out)
    model.ipopt_zU_in.update(model.ipopt_zU_out)

    return None



"""
This module contains the main methods used for reading, writing, and manipulating
data.

The read_data and write_data methods are accessible at the top level (i.e. kipet.read_data)
"""
# Standard library imports
import os
import pathlib
import re
import sys
import traceback

# Third party imports
import numpy as np
import pandas as pd


# KIPET library imports
# from kipet.model_tools.pyomo_model_tools import _df_from_pyomo_data

# Public methods that can be used in KIPET
__all__ = ['write_data', 'read_data', 'read_file', 'add_noise_to_data', 'Tee']


# Context manager that copies stdout and any exceptions to a log file
class Tee:
    
    def __init__(self, filename):
    
        self.file = open(filename, 'a+')
        self.stdout = sys.stdout
        #self.stderr = sys.stderr
    
    def __enter__(self):
    
        sys.stdout = self
    
    def __exit__(self, exc_type, exc_value, tb):
    
        sys.stdout = self.stdout
        if exc_type is not None:
            self.file.write(traceback.format_exc())
        self.file.close()
    
    def write(self, data):
        
        self.file.write(data)
        self.stdout.write(data)
    
    def flush(self):
        
        self.file.flush()
        self.stdout.flush()

def write_data(filename, data):
    """Method to write data to a file using KIPET

    Convenient method to save modified data in a format ready to use with ReactionModels.

    :parameter str filename: The name of the file (plus relative directory)
    :parameter pandas.DataFrame data: The pandas DataFrame to be written to the file

    :returns: None
    
    """
    calling_file_name = os.path.dirname(os.path.realpath(sys.argv[0]))
    write_file(pathlib.Path(calling_file_name).joinpath(filename), data)

    return None


def read_data(filename):
    """Method to read data file using KipetModel
    
    This is useful if you need to modify a datafile before using it with a ReactionModel.

    :parameter str filename: The name of the data file (expected to be in the data
       directory, otherwise use an absolute path).

    :return loaded_data: The data read from the file
    :rtype: pandas.DataFrame
    
    """
    _filename = filename
    calling_file_name = os.path.dirname(os.path.realpath(sys.argv[0]))
    _filename = pathlib.Path(calling_file_name).joinpath(_filename)
    loaded_data = read_file(_filename)

    return loaded_data


def read_file(filename):
    """ Reads data from a csv or txt file and converts it to a DataFrame

        :param str filename: name of input file (abs path)
          
        :return df_data: DataFrame
        :rtype: pandas.DataFrame

        :Raises:

            ValueError if the file type is not CSV or TXT

    """
    filename = pathlib.Path(filename)
    data_dict = {}

    if filename.suffix == '.txt':
        with open(filename, 'r') as f:
            for line in f:
                if line not in ['', '\n', '\t', '\t\n']:
                    l = line.split()
                    if is_float_re(l[1]):
                        l[1] = float(l[1])
                    data_dict[float(l[0]), l[1]] = float(l[2])

        df_data = dict_to_df(data_dict)
        df_data.sort_index(ascending=True, inplace=True)

    elif filename.suffix == '.csv':
        df_data = pd.read_csv(filename, index_col=0)

    else:
        raise ValueError(f'The file extension {filename.suffix} is currently not supported')
        return None

    return df_data


def write_file(filename, dataframe, filetype='csv'):
    """ Write data from a dataframe to file.
    
        :param str filename: Name of output file (abs_path)
        :param pandas.DataFrame dataframe: Data to write to file
        :param str filetype: Choice of file output (CSV, TXT)
        
        :return: None

    """
    if filetype not in ['csv', 'txt']:
        print('Savings as CSV - invalid file extension given')
        filetype = 'csv'

    suffix = '.' + filetype

    filename = pathlib.Path(filename)

    if filename.suffix == '':
        filename = filename.with_suffix(suffix)
    else:
        suffix = filename.suffix
        if suffix not in ['.txt', '.csv']:
            print('Savings as CSV - invalid file extension given')
            filename = pathlib.Path(filename.stem).with_suffix('.csv')

    print(f'In write method: {filename.absolute()}')
    if filename.suffix == '.csv':
        dataframe.to_csv(filename)

    elif filename.suffix == '.txt':
        with open(filename, 'w') as f:
            for i in dataframe.index:
                for j in dataframe.columns:
                    if not np.isnan(dataframe[j][i]):
                        f.write("{0} {1} {2}\n".format(i, j, dataframe[j][i]))

    print(f'Data saved     : {filename}')
    return None


def read_spectral_data_from_csv(filename, instrument=False, negatives_to_zero=False):
    """ Reads csv with spectral data

        :param str filename: Name of input file
        :param bool instrument: Indicates if data is direct from instrument
        :param bool negatives_to_zero: If data contains negatives and baseline shift is not
                                        done then this forces negative values to zero.

        :return data: The dataframe of spectral data after formatting
        :rtype: pandas.DataFrame

    """
    data = pd.read_csv(filename, index_col=0)
    if instrument:
        # this means we probably have a date/timestamp on the columns
        data = pd.read_csv(filename, index_col=0, parse_dates=True)
        data = data.T
        for n in data.index:
            h, m, s = n.split(':')
            sec = (float(h) * 60 + float(m)) * 60 + float(s)
            data.rename(index={n: sec}, inplace=True)
        data.index = [float(n) for n in data.index]
    else:
        data.columns = [float(n) for n in data.columns]

    # If we have negative values then this makes them equal to zero
    if negatives_to_zero:
        data[data < 0] = 0

    return data


def is_float_re(str_var):
    """Checks if a value is a float or not

    :param str str_var: Checks if the string value is a float

    :return: boolean indicating if the string can be considered a float
    :rtype: bool

    """
    _float_regexp = re.compile(r"^[-+]?(?:\b[0-9]+(?:\.[0-9]*)?|\.[0-9]+\b)(?:[eE][-+]?[0-9]+\b)?$").match
    return True if _float_regexp(str_var) else False


def dict_to_df(data_dict):
    """Takes a dictionary of typical pyomo data and converts it to a dataframe

    :param dict data_dict: dict of the Pyomo data

    :return dfs: DataFrame of the unpacked data
    :rtype: pandas.DataFrame
    
    """
    dfs_stacked = pd.Series(index=data_dict.keys(), data=list(data_dict.values()))
    dfs = dfs_stacked.unstack()
    return dfs


def add_noise_to_data(data, noise):
    """Wrapper for adding noise to data after data has been added to
    the specific ReactionModel
    
    :parameter pandas.DataFrame data: The dataset to which noise is to be added.
    :parameter float noise: The variance of the added noise.
    
    :return noised_data: The dataset after noised has been added.
    :rtype: pandas.DataFrame

    """
    from kipet.calculation_tools.prob_gen_tools import add_noise_to_signal
    noised_data = add_noise_to_signal(data, noise)

    return noised_data



"""
Unit conversion helper functions
"""
# Standard library imports
from enum import Enum

# Third party imports
import attr


class ConversionType(Enum):
    """Simple Enum class for conversion types"""

    volume = 1
    time = 2
    generic = 3


@attr.s
class UnitConversion:
    """Unit conversion class

    :param str unit: The units considered
    :param str dim: The dimensions of the unit
    :param int power: The power of the unit
    :param int dim_power: The power of the dimension

    """
    unit = attr.ib()
    dim = attr.ib()
    power = attr.ib()
    dim_power = attr.ib()


def convert_unit(unit_registry, u_orig, u_goal, scalar=1, power=1, both_powers=False, reverse_power=False):
    """Method to convert units taking the powers into account

    :param unit_registry: The unit registry to use (must be the same)
    :param str u_orig: The original units
    :param str u_goal: The target units to convert to
    :param float scalar: The scalar value to pass to the new units (usually 1)
    :param int power: The power of the unit dimension
    :param bool both_powers: Option to use both powers or a single unit's power
    :param bool reverse_power: Option to reverse the power (invert)

    :return: The new units
    :rtype: pint.Quantity

    """
    c1 = 1*unit_registry(u_orig)
    c2 = 1*unit_registry(u_goal)

    power2 = 1
    if both_powers:
        power2 = power

    if reverse_power:
        power2, power = power, power2

    con = (c1**power).to((c2**power2).u)/c2**power2 * (c2**power2).u/(c1**power).u

    return scalar * con


def convert_single_dimension(unit_registry, unit_from, unit_to):
    """This method converts only a single dimension from the total unit

    :param unit_registry: The required unit registry object
    :param str unit_from: The unit to be converted
    :param str unit_to: The target unit

    :return new_unit: The unit with the updated dimension units
    :rtype: pint.Quantity

    """
    conversion = ConversionType.generic

    if isinstance(unit_to, str):
        unit_to = unit_registry(unit_to)
        is_volume = unit_to.check('[volume]')

    if isinstance(unit_from, str):
        unit_from = unit_registry(unit_from)

    if is_volume:
        conversion = ConversionType.volume

    u_to = unit_to
    u_from = unit_from

    dim_to = {k.replace('[', '').replace(']', ''): v for k, v in dict(u_to.dimensionality).items()}
    dim_from = {k.replace('[', '').replace(']', ''): v for k, v in dict(u_from.dimensionality).items()}

    units_to = {k: v for k, v in u_to._units.items()}
    units_from = {k: v for k, v in u_from._units.items()}
    new_unit = unit_from

    for unit in units_from:
        dims = dict(unit_registry(unit).dimensionality)
        dims = {k.replace('[', '').replace(']', ''): v for k, v in dims.items()}

        try:
            uc_from = UnitConversion(unit=unit,
                                 dim=list(dims.keys())[0],
                                 power=units_from[unit],
                                 dim_power=dim_from[list(dims.keys())[0]]
                                 )
        finally:
            continue

        if (unit_registry(uc_from.unit)**uc_from.power).check("[power]"):
            continue

        if uc_from.unit == uc_to.unit:
            _print('The units are already equal...quitting')
            break

        if uc_from.dim == uc_to.dim:

            if conversion != ConversionType.volume:
                is_volume = (unit_registry(uc_from.unit)**uc_from.power).check("[volume]")
                if is_volume:
                    continue

                reverse_power = False
                if uc_from.power > uc_to.power:
                    reverse_power = True

                power = uc_from.power
                con = convert_unit(unit_registry, unit, uc_to.unit, power=abs(power), reverse_power=reverse_power)
                con = con ** (power/abs(power))
                new_unit = unit_from * con

            elif conversion == ConversionType.volume:
                if unit_registry(f'{unit}**{abs(uc_from.power)}').check('[volume]'):
                    power = 3
                    if abs(uc_from.power) == abs(uc_to.power):
                        power = uc_from.power

                    reverse_power = False
                    if abs(uc_from.power) < abs(uc_to.power):
                        reverse_power = True

                    both = False
                    if abs(uc_from.power) == abs(uc_to.power):
                        both = True

                    con = convert_unit(unit_registry, unit, uc_to.unit, power=power, both_powers=both, reverse_power=reverse_power)
                    con = con ** (uc_from.power/abs(uc_from.power))
                    new_unit = unit_from * con

                else:
                    continue

    return new_unit


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Jun 22 15:58:34 2021

@author: kevinmcbride
"""

  # Make simulation results go somewhere else - don't overwrite the existing results
    
    def _plot_all_Z_bounds(self):
        """Plot all concentration profiles
        
        """
        #%%
        # self = r1._plot_object
        # self.show = True
        # import numpy as np
        
        # fig = go.Figure()
        # use_spectral_format = False
        # pred = getattr(self.reaction_model.results, 'Z')
        # if hasattr(self.reaction_model.results, 'Cm'):
        #     exp = getattr(self.reaction_model.results, 'Cm')
        # elif hasattr(self.reaction_model.results, 'C'):
        #     if self.reaction_model.models['_s_model'] and not self.reaction_model.models['v_model'] and not self.reaction_model.models['p_model']:
        #         exp = None
        #     else:
        #         exp = getattr(self.reaction_model.results, 'C')
        #         use_spectral_format = True
        # else:
        #     exp = None
            
        # pred = self.reaction_model.results.Z
            
        # p_plus = {k: self.reaction_model.results_dict['p_estimator'].P[k] + self.reaction_model.results_dict['p_estimator'].deviations(0.99999)[k] for k in self.reaction_model.results_dict['p_estimator'].P.keys()}
        # self.reaction_model.simulate(parameters=p_plus)
        # pred_plus = self.reaction_model.results_dict['simulator'].Z
        
        # p_minus = {k: self.reaction_model.results_dict['p_estimator'].P[k] - self.reaction_model.results_dict['p_estimator'].deviations(0.99999)[k] for k in self.reaction_model.results_dict['p_estimator'].P.keys()}
        # self.reaction_model.simulate(parameters=p_minus)
        # pred_minus = self.reaction_model.results_dict['simulator'].Z

        # for i, col in enumerate(pred.columns):
            
        #     self._state_plot(fig, col, pred, exp, use_spectral_format=use_spectral_format)
            
        #     fig.add_trace(go.Scatter(
        #         x=np.concatenate([pred.index, pred.index[::-1]]),
        #         y=pd.concat([pred_plus[col], pred_minus[::-1][col]]),
        #         fill='toself',
        #         hoveron='points',
        #         opacity=0.2,
        #         fillcolor=colors[self.color_num],
        #         line=dict(color=colors[self.color_num], width=0),
        #     ))
        #     self.color_num += 1
        # # fig.show()
        # pio.write_html(fig, file='test', auto_open=True)    
        #%%
        
        # self.color_num = 0
        # var_data = self.reaction_model.components[col]
        # state = f'{self.reaction_model.components[col].state}'.capitalize()
        # title = f'Model: {self.reaction_model.name} | Concentration Profiles'
        # time_scale = f'Time [{self.reaction_model.unit_base.time}]'

        # state_units = self._get_proper_unit_str(var_data)
        # # state_units = var_data.units.u
        # fig.update_layout(
        #         title=title,
        #         xaxis_title=f'{time_scale}',
        #         yaxis_title=f'{state} [{state_units}]',
        #         )
        # self._fig_finishing(fig, pred, plot_name='all-concentration-profiles')

        #%%

        return None
    
#%%



"""
FESimulator - sets up the model for use with fe_factory
"""
# KIPET library imports
from kipet.estimator_tools.fe_factory import FEInitialize
from kipet.estimator_tools.pyomo_simulator import PyomoSimulator
from kipet.general_settings.variable_names import VariableNames

__author__ = "Michael Short, Kevin McBride"


class FESimulator(PyomoSimulator):
    """This class is just an interface that allows Kipet to easily
    implement the more general fe_factory class designed by David M. Thierry 
    without having to re-write the model to fit those arguments.
    
    """

    def __init__(self, model):
        """It takes in a standard Kipet/Pyomo model, rewrites it and calls 
        fe_factory. More information on fe_factory is included in that class
        description.

        :param ConcreteModel model: Pyomo model passed from KIPET
        
        """
        super(FESimulator, self).__init__(model)
        self.__var = VariableNames()
        self.p_sim = PyomoSimulator(model)
        self.model = self.p_sim.model
        self.param_dict = {}
        self.param_name = self.__var.model_parameter
        self.c_sim = self.model.clone()

        # Check all parameters are fixed before simulating
        if hasattr(self.model, self.__var.model_parameter):
            for param_obj in getattr(self.model, self.__var.model_parameter).values():
                if not param_obj.fixed:
                    param_obj.fixed = True
            # Build the parameter dictionary in the format that fe_factory uses    
            model_var = self.__var.model_parameter

            for k, v in getattr(self.model, model_var).items():
                self.param_dict[model_var, k] = v.value

        # Build the initial condition dictionary in the format that fe_factory uses
        vars_to_init = [self.__var.concentration_model,
                        self.__var.state_model,
                        ]

        self.ics_ = {}
        for var in vars_to_init:
            self._var_init(var)

    def _var_init(self, model_var):
        """Initializes the IC dict for the state variables and modifies it in
        place.

        :param str model_var: The model variable to be used in initializing the initial condition dict
            
        :return: None

        """
        if hasattr(self.model, model_var):
            for k, v in getattr(self.model, model_var).items():
                st = self.model.start_time
                if k[0] == st:
                    self.ics_[model_var, k[1]] = v.value

        return None

    def call_fe_factory(self, inputs_sub=None, dosing_points=None):
        """This function applies all the inputs necessary for fe_factory to
        work, using Kipet syntax. Requires external inputs/dosing points to be
        specified with the following arguments.


        :param dict inputs_sub: dictionary of inputs
        :param dict dosing_points: dictionary of the dosing points
            
        :return: None
        
        """
        self.inputs_sub = inputs_sub

        init = FEInitialize(self.model,
                             self.c_sim,
                             init_con="init_conditions_c",
                             param_name=self.param_name,
                             param_values=self.param_dict,
                             inputs_sub=self.inputs_sub,
                             )

        init.load_initial_conditions(init_cond=self.ics_)

        if dosing_points is not None:
            init.load_discrete_jump(dosing_points)

        init.run()

        return None
